import os
import boto3

from botocore.config import Config
from botocore.exceptions import ClientError
from elasticsearch.exceptions import (
    ConflictError,
    ConnectionError,
    NotFoundError,
    TransportError,
)
from pyramid.view import view_config
from pyramid.settings import asbool
from sqlalchemy.exc import StatementError
from snovault import (
    COLLECTIONS,
    DBSESSION,
    STORAGE
)
from snovault.storage import (
    TransactionRecord,
)
from urllib3.exceptions import ReadTimeoutError
from .interfaces import (
    ELASTIC_SEARCH,
    INDEXER,
    RESOURCES_INDEX,
)
from .indexer_state import (
    IndexerState,
    all_uuids,
    all_types,
    SEARCH_MAX,
    setup_indexing_nodes,
    HEAD_NODE_INDEX,
    INDEXING_NODE_INDEX,
    AWS_REGION,
)
from .simple_queue import SimpleUuidServer

import datetime
import logging
import pytz
import time
import copy
import json
import requests

es_logger = logging.getLogger("elasticsearch")
es_logger.setLevel(logging.ERROR)
log = logging.getLogger('snovault.elasticsearch.es_index_listener')
MAX_CLAUSES_FOR_ES = 8192
DEFAULT_QUEUE = 'Simple'
# Allow three minutes for indexer to stop before head indexer process continues
_REMOTE_INDEXING_SHUTDOWN_SLEEP = 3*60
# Allow indexer to remain up for 30 minutes after indexing before being shutdown
_REMOTE_INDEXING_SHUTDOWN = 15*60
# Allow indexer 20 minutes to startup before head indexer thread takes over indexing
_REMOTE_INDEXING_STARTUP = 10*60
# The number of uuids needed to start a remote indexer
_REMOTE_INDEXING_THRESHOLD = 10000


def _update_for_uuid_queues(registry):
    """
    Update registry with uuid queue module if it exists
    """
    extra_queues = []
    try:
        import snovault.elasticsearch.uuid_queue as queue_adapter
    except ImportError as ecp:
        log.info('No uuid_queue package in elasticsearch module: %s', repr(ecp))
    else:
        registry['UuidQueue'] = queue_adapter.QueueAdapter
        extra_queues = queue_adapter.QueueTypes.get_all()
        log.info('Extra Indexer Queues Available: %s', ','.join(extra_queues))
    registry['available_queues'].extend(extra_queues)


def includeme(config):
    """Add index listener endpoint and setup Indexer"""
    config.add_route('index', '/index')
    config.scan(__name__)
    registry = config.registry
    processes = registry.settings.get('indexer.processes')
    is_indexer = registry.settings.get('indexer')
    if is_indexer:
        available_queues = [DEFAULT_QUEUE]
        registry['available_queues'] = available_queues
        _update_for_uuid_queues(registry)
        if not processes:
            registry[INDEXER] = Indexer(registry)


def get_related_uuids(request, es, updated, renamed):
    '''Returns (set of uuids, False) or (list of all uuids, True) if full reindex triggered'''

    updated_count = len(updated)
    renamed_count = len(renamed)
    if (updated_count + renamed_count) > MAX_CLAUSES_FOR_ES:
        return (list(all_uuids(request.registry)), True)  # guaranteed unique
    elif (updated_count + renamed_count) == 0:
        return (set(), False)

    es.indices.refresh(RESOURCES_INDEX)

    # TODO: batching may allow us to drive a partial reindexing much greater than 99999
    #BATCH_COUNT = 100  # NOTE: 100 random uuids returned > 99999 results!
    #beg = 0
    #end = BATCH_COUNT
    #related_set = set()
    #updated_list = list(updated)  # Must be lists
    #renamed_list = list(renamed)
    #while updated_count > beg or renamed_count > beg:
    #    if updated_count > end or beg > 0:
    #        log.error('Indexer looking for related uuids by BATCH[%d,%d]' % (beg, end))
    #
    #    updated = []
    #    if updated_count > beg:
    #        updated = updated_list[beg:end]
    #    renamed = []
    #    if renamed_count > beg:
    #        renamed = renamed_list[beg:end]
    #
    #     search ...
    #     accumulate...
    #
    #    beg += BATCH_COUNT
    #    end += BATCH_COUNT

    query = {
        'query': {
            'bool': {
                'should': [
                    {
                        'terms': {
                            'embedded_uuids': updated,
                            '_cache': False,
                        },
                    },
                    {
                        'terms': {
                            'linked_uuids': renamed,
                            '_cache': False,
                        },
                    },
                ],
            },
        },
        '_source': False,
    }
    res = es.search(index=RESOURCES_INDEX, size=SEARCH_MAX, request_timeout=60, body=query)

    if res['hits']['total'] > SEARCH_MAX:
        return (list(all_uuids(request.registry)), True)  # guaranteed unique

    related_set = {hit['_id'] for hit in res['hits']['hits']}

    return (related_set, False)


def _determine_indexing_protocol(request, uuid_count):
    remote_indexing = asbool(
        os.environ.get("REMOTE_INDEXING")
        or request.registry.settings.get('remote_indexing', False)
    )
    if not remote_indexing:
        return False
    try:
        indexer_short_uuids = int(request.registry.settings.get('indexer_short_uuids', 0))
        remote_indexing_threshold = int(
            request.registry.settings.get(
                'remote_indexing_threshold',
                _REMOTE_INDEXING_THRESHOLD,
            )
        )
        if indexer_short_uuids > 0:
            uuid_count = indexer_short_uuids
    except ValueError:
        log.warning('ValueError casting remote indexing threshold to an int')
        remote_indexing_threshold = _REMOTE_INDEXING_THRESHOLD
    if uuid_count < remote_indexing_threshold:
        return False
    return True


def _send_sqs_msg(
            cluster_name,
            remote_indexing,
            body,
            delay_seconds=0,
            save_event=False,
        ):
    did_fail = True
    msg_attrs = {
        'cluster_name': cluster_name,
        'remote_indexing': '0' if remote_indexing else '1',
        'save_event': '0' if save_event else '1',
    }
    msg_attrs = {
        key: {'DataType': 'String', 'StringValue': val}
        for key, val in msg_attrs.items()
    }
    boto_session = boto3.Session(region_name=AWS_REGION)
    sqs_resource = boto_session.resource('sqs')
    watch_dog = 3
    attempts = 0
    while attempts < watch_dog:
        attempts += 1
        try: 
            queue_arn = sqs_resource.get_queue_by_name(QueueName='ec2_scheduler_input')
            if queue_arn:
                response = sqs_resource.meta.client.send_message(
                    QueueUrl=queue_arn.url,
                    DelaySeconds=delay_seconds,
                    MessageAttributes=msg_attrs,
                    MessageBody=body,
                )
                if response.get('ResponseMetadata', {}).get('HTTPStatusCode') == 200:
                    log.warning(f"Remote indexing _send_sqs_msg: {body}")
                    did_fail = False
                    break
        except ClientError as ecp:
            error_str = e.response.get('Error', {}).get('Code', 'Unknown')
            log.warning(f"({attempts} of {watch_dog}) Indexer could not send message to queue: {error_str}")
    return did_fail


def _get_nodes(request, indexer_state):
    label = ''
    continue_on = False
    did_timeout = False
    # Get local/remote indexing state
    head_node, indexing_node, time_now, did_fail, is_indexing_node = setup_indexing_nodes(
        request, indexer_state
    )
    this_node = head_node
    other_node = indexing_node
    if is_indexing_node:
        this_node = indexing_node
        other_node = head_node
    if did_fail:
        did_timeout = True
        return this_node, other_node, continue_on, did_timeout
    this_time_in_state = time_now - float(this_node['last_run_time'])
    other_time_in_state = time_now - float(other_node['last_run_time'])
    # Check for early return if indexing is running
    if this_node['node_index'] == HEAD_NODE_INDEX:
        if other_node['done_indexing']:
            label = 'Head node with remote finished indexing.  Reset both es node states.'
            this_node, other_node, time_now, did_fail, is_indexing_node = setup_indexing_nodes(
                request,
                indexer_state,
                reset=True
            )
        elif other_node['started_indexing']:
            # not catching if instance is on but app not running?  maybe the indexer state
            #  needs update a flag saying it is still running
            if other_node['instance_state'] in ['running', 'pending']:
                # label = 'Head node with remote indexing running'
                pass
            else:
                label = 'Head node with remote indexing running, permature shutdown'
                log.warning(f"Indexer permature shutdown: sleep {_REMOTE_INDEXING_SHUTDOWN_SLEEP} seconds")
                time.sleep(_REMOTE_INDEXING_SHUTDOWN_SLEEP)
                continue_on = True
        elif this_node['waiting_on_remote']:
            label = 'Head node waiting on remote indexing to start'
            if this_time_in_state > _REMOTE_INDEXING_STARTUP:
                this_node, other_node, time_now, did_fail, is_indexing_node = setup_indexing_nodes(
                    request,
                    indexer_state,
                    reset=True
                )
                if not other_node['instance_state'] in ['stopped', 'stopping']:
                    instance_name = this_node['instance_name']
                    remote_indexing = False
                    _ = _send_sqs_msg(
                        f"{instance_name}-indexer",
                        remote_indexing,
                        'Turn OFF remote indexer for failed start',
                        delay_seconds=0,
                        save_event=False
                    )
                    log.warning(f"Indexer failed start shutdown sleep: {_REMOTE_INDEXING_SHUTDOWN_SLEEP} seconds")
                    time.sleep(_REMOTE_INDEXING_SHUTDOWN_SLEEP)
                else:
                    # log.warning('Indexer failed start shutdown: Already stopped')
                    pass
                did_timeout = True
        else:
            # label = 'Head node says no indexing going on'
            if this_time_in_state > _REMOTE_INDEXING_SHUTDOWN:
                this_node, other_node, time_now, did_fail, is_indexing_node = setup_indexing_nodes(
                    request,
                    indexer_state,
                    reset=True
                )
                if not other_node['instance_state'] in ['stopped', 'stopping']:
                    label = 'Head node says no indexing going on - and shutting down indexer'
                    instance_name = this_node['instance_name']
                    remote_indexing = False
                    _ = _send_sqs_msg(
                        f"{instance_name}-indexer",
                        remote_indexing,
                        'Turn OFF remote indexer for shutdown',
                        delay_seconds=0,
                        save_event=False
                    )
                    log.warning(f"Indexer timeout shutdown: sleep {_REMOTE_INDEXING_SHUTDOWN_SLEEP} seconds")
                    time.sleep(_REMOTE_INDEXING_SHUTDOWN_SLEEP)
                else:
                    # log.warning('Indexer timeout shutdown: Already stopped')
                    pass
            continue_on = True
    elif this_node['node_index'] == INDEXING_NODE_INDEX:
        if other_node['waiting_on_remote']:
            if not this_node['started_indexing']:
                label = 'Index node received signal to start indexing'
                continue_on = True
            elif this_node['done_indexing']:
                label = 'Index node finished indexing.  Waiting on head node to pick up status.'
            elif this_node['started_indexing']:
                label = 'Index node has just finished indexing'
                this_node['done_indexing'] = True
                this_node['last_run_time'] = str(float(time.time()))
                indexer_state.put_obj(this_node['node_index'], this_node)
            else:
                label = 'Index node is in limbo with head node waiting_on_remote.  May correct on next loop.'
        else:
            label = 'Index node finished indexing and has been reset by head node.  Waiting to be shutdown by head node.'
    if label:
        if other_node['instance_state'] == 'stopped':
            log.warning(f"Remote indexing _get_nodes: {label}. this_time: {this_time_in_state:0.6f}, other_time: stopped")
        else:
            log.warning(f"Remote indexing _get_nodes: {label}. this_time: {this_time_in_state:0.6f}, other_time: {other_time_in_state:0.6f}")
    return this_node, other_node, continue_on, did_timeout


@view_config(route_name='index', request_method='POST', permission="index")
def index(request):
    # Setting request.datastore here only works because routed views are not traversed.
    request.datastore = 'database'
    session = request.registry[DBSESSION]()
    connection = session.connection()
    # Currently 2 possible followup indexers (base.ini [set stage_for_followup = vis_indexer, region_indexer])
    stage_for_followup = list(request.registry.settings.get("stage_for_followup", '').replace(' ','').split(','))
    # May have undone uuids from prior cycle
    indexer_state = IndexerState(
        request.registry[ELASTIC_SEARCH],
        request.registry.settings['snovault.elasticsearch.index'],
        followups=stage_for_followup,
    )
    # Check remote indexing and set head/indexing nodes
    this_node = indexer_state.get_obj(INDEXING_NODE_INDEX)
    other_node = indexer_state.get_obj(HEAD_NODE_INDEX)
    remote_indexing = asbool(
        os.environ.get("REMOTE_INDEXING")
        or request.registry.settings.get('remote_indexing', False)
    )
    did_timeout = False
    if remote_indexing:
        this_node, other_node, continue_on, did_timeout = _get_nodes(request, indexer_state)
        if not did_timeout and not continue_on:
            # Node state implies we do not need to run any code right now
            return {'this_node': this_node, 'other_node': other_node}
    elif this_node and other_node:
        # if remote node states exists and remote_indexing false then this is the remote indexer
        this_node, other_node, continue_on, did_timeout = _get_nodes(request, indexer_state)
        if not did_timeout and not continue_on:
            # Node state implies we do not need to run any code right now
            return {'this_node': this_node, 'other_node': other_node}

    def _load_indexing(request, session, connection, indexer_state):
        first_txn = None
        snapshot_id = None
        (xmin, invalidated, restart) = indexer_state.priority_cycle(request)
        indexer_state.log_reindex_init_state()
        # OPTIONAL: restart support
        if restart:  # Currently not bothering with restart!!!
            xmin = -1
            invalidated = []
        # OPTIONAL: restart support

        result = indexer_state.get_initial_state()  # get after checking priority!

        if xmin == -1 or len(invalidated) == 0:
            xmin = get_current_xmin(request)

            last_xmin = None
            if 'last_xmin' in request.json:
                last_xmin = request.json['last_xmin']
            else:
                status = request.registry[ELASTIC_SEARCH].get(
                    index=request.registry.settings['snovault.elasticsearch.index'],
                    doc_type='meta',
                    id='indexing',
                    ignore=[400, 404]
                )
                if status['found'] and 'xmin' in status['_source']:
                    last_xmin = status['_source']['xmin']
            if last_xmin is None:  # still!
                if 'last_xmin' in result:
                    last_xmin = result['last_xmin']
                elif 'xmin' in result and result['xmin'] < xmin:
                    last_xmin = result['state']

            result.update(
                xmin=xmin,
                last_xmin=last_xmin,
            )

        if len(invalidated) > SEARCH_MAX:  # Priority cycle already set up
            flush = True
        else:

            flush = False
            if last_xmin is None:
                result['types'] = types = request.json.get('types', None)
                invalidated = list(all_uuids(request.registry, types))
                flush = True
            else:
                txns = session.query(TransactionRecord).filter(
                    TransactionRecord.xid >= last_xmin,
                )

                invalidated = set(invalidated)  # not empty if API index request occurred
                updated = set()
                renamed = set()
                max_xid = 0
                txn_count = 0
                for txn in txns.all():
                    txn_count += 1
                    max_xid = max(max_xid, txn.xid)
                    if first_txn is None:
                        first_txn = txn.timestamp
                    else:
                        first_txn = min(first_txn, txn.timestamp)
                    renamed.update(txn.data.get('renamed', ()))
                    updated.update(txn.data.get('updated', ()))

                if invalidated:        # reindex requested, treat like updated
                    updated |= invalidated

                result['txn_count'] = txn_count
                if txn_count == 0 and len(invalidated) == 0:
                    indexer_state.send_notices()
                    return_now = True
                    return result, return_now, flush, invalidated, first_txn, snapshot_id, restart, xmin

                is_testing = asbool(request.registry.settings.get('testing', False))
                is_testing_full = request.json.get('is_testing_full', False)
                if is_testing and is_testing_full:
                    full_reindex = False
                    related_set = set(all_uuids(request.registry))
                else:
                    (related_set, full_reindex) = get_related_uuids(request, request.registry[ELASTIC_SEARCH], updated, renamed)
                if full_reindex:
                    invalidated = related_set
                    flush = True
                else:
                    invalidated = related_set | updated
                    result.update(
                        max_xid=max_xid,
                        renamed=renamed,
                        updated=updated,
                        referencing=len(related_set),
                        invalidated=len(invalidated),
                        txn_count=txn_count
                    )
                    if first_txn is not None:
                        result['first_txn_timestamp'] = first_txn.isoformat()

                if invalidated and not request.json.get('dry_run', False):
                    # Exporting a snapshot mints a new xid, so only do so when required.
                    # Not yet possible to export a snapshot on a standby server:
                    # http://www.postgresql.org/message-id/CAHGQGwEtJCeHUB6KzaiJ6ndvx6EFsidTGnuLwJ1itwVH0EJTOA@mail.gmail.com
                    if snapshot_id is None and not request.json.get('recovery', False):
                        snapshot_id = connection.execute('SELECT pg_export_snapshot();').scalar()
        return_now = False
        return result, invalidated, flush, first_txn, snapshot_id, restart, xmin, return_now


    def _run_indexing(
            request,
            indexer_state,
            result,
            invalidated,
            stage_for_followup,
            flush,
            snapshot_id,
            restart,
            xmin,
        ):
        if len(stage_for_followup) > 0:
            # Note: undones should be added before, because those uuids will (hopefully) be indexed in this cycle
            indexer_state.prep_for_followup(xmin, invalidated)

        result = indexer_state.start_cycle(invalidated, result)

        # Do the work...

        indexing_update_infos, errors, err_msg = request.registry[INDEXER].serve_objects(
            request,
            invalidated,
            xmin,
            snapshot_id=snapshot_id,
            restart=restart,
        )
        if err_msg:
            log.warning('Could not start indexing: %s', err_msg)
        result = indexer_state.finish_cycle(result,errors)

        if errors:
            result['errors'] = errors

        if request.json.get('record', False):
            try:
                request.registry[ELASTIC_SEARCH].index(
                    index=request.registry.settings['snovault.elasticsearch.index'],
                    doc_type='meta',
                    body=result,
                    id='indexing'
                )
            except:
                error_messages = copy.deepcopy(result['errors'])
                del result['errors']
                request.registry[ELASTIC_SEARCH].index(
                    index=request.registry.settings['snovault.elasticsearch.index'],
                    doc_type='meta',
                    body=result,
                    id='indexing'
                )
                for item in error_messages:
                    if 'error_message' in item:
                        log.error('Indexing error for {}, error message: {}'.format(item['uuid'], item['error_message']))
                        item['error_message'] = "Error occured during indexing, check the logs"
                result['errors'] = error_messages


        request.registry[ELASTIC_SEARCH].indices.refresh(RESOURCES_INDEX)
        if flush:
            try:
                request.registry[ELASTIC_SEARCH].indices.flush_synced(index=RESOURCES_INDEX)  # Faster recovery on ES restart
            except ConflictError:
                pass
        return result, indexing_update_infos

    # Load indexing
    output_tuple = _load_indexing(request, session, connection, indexer_state)
    result, invalidated, flush, first_txn, snapshot_id, restart, xmin, return_now = output_tuple
    if return_now:
        return result
    indexing_update_infos = []
    dry_run = request.json.get('dry_run', False)
    if invalidated and not dry_run:
        run_indexing_vars = (
            request,
            indexer_state,
            result,
            invalidated,
            stage_for_followup,
            flush,
            snapshot_id,
            restart,
            xmin,
        )
        if this_node:
            if this_node['node_index'] == HEAD_NODE_INDEX:
                # Check remote indexing
                uuids_count = len(invalidated)
                do_remote_indexing = _determine_indexing_protocol(request, uuids_count)
                if did_timeout:
                    log.warning(f"Remote indexing index: did_timeout.  Run on head node.")
                    head_node, indexing_node, time_now, did_fail, is_indexing_node  = setup_indexing_nodes(
                        request,
                        indexer_state,
                        reset=True
                    )
                    result, indexing_update_infos = _run_indexing(*run_indexing_vars)
                elif do_remote_indexing:
                    log.warning(f"Remote indexing run indexing: Defer to remote.")
                    instance_name = this_node['instance_name']
                    did_fail = False
                    if not other_node['instance_state'] in ['running', 'pending']:
                        did_fail = _send_sqs_msg(
                            f"{instance_name}-indexer",
                            remote_indexing,
                            'Turn ON remote indexer',
                            delay_seconds=0,
                            save_event=False
                        )
                    else:
                        log.warning(f"Remote indexing run indexing: Already on.")
                    if not did_fail:
                        this_node['waiting_on_remote'] = True
                        this_node['last_run_time'] = str(float(time.time()))
                        indexer_state.put_obj(this_node['node_index'], this_node)
                    else:
                        log.warning(f"Remote indexing index: Failed sqs queue.  Run on head node.")
                        result, indexing_update_infos = _run_indexing(*run_indexing_vars)
                else:
                    log.warning(f"Remote indexing index: Below threshold. Run on head node.")
                    result, indexing_update_infos = _run_indexing(*run_indexing_vars)
                    # Do not reset state for local indexing below threshold.
            elif this_node['node_index'] == INDEXING_NODE_INDEX:
                log.warning(f"Remote indexing indexing node: Start indexing")
                this_node['started_indexing'] = True
                this_node['last_run_time'] = str(float(time.time()))
                indexer_state.put_obj(this_node['node_index'], this_node)
                result, indexing_update_infos = _run_indexing(*run_indexing_vars)
                this_node['done_indexing'] = True
                this_node['last_run_time'] = str(float(time.time()))
                indexer_state.put_obj(this_node['node_index'], this_node)
                log.warning(f"Remote indexing indexing node: Done indexing")
        else:
            log.warning(f"Non remote indexing")
            result, indexing_update_infos = _run_indexing(*run_indexing_vars)
    elif invalidated:
        log.warning(f"Load indexing have uuids but dry_run")
    elif dry_run:
        log.warning(f"Load indexing no uuids on dry_run")
    # Add transaction lag to result
    if first_txn is not None:
        result['txn_lag'] = str(datetime.datetime.now(pytz.utc) - first_txn)
    # Send slack notifications
    indexer_state.send_notices()
    # Log indexing times
    if indexing_update_infos:
        # Check for logging of intitial indexing info here,
        #  opposed to in the indexer or just after serve_objects,
        #  so a crash in logging does not interupt indexing complietion
        request.registry[INDEXER].check_log_indexing_times(indexing_update_infos)
    return result


def get_current_xmin(request):
    session = request.registry[DBSESSION]()
    connection = session.connection()
    recovery = request.json.get('recovery', False)

    # http://www.postgresql.org/docs/9.3/static/functions-info.html#FUNCTIONS-TXID-SNAPSHOT
    if recovery:
        query = connection.execute(
            "SET TRANSACTION ISOLATION LEVEL READ COMMITTED, READ ONLY;"
            "SELECT txid_snapshot_xmin(txid_current_snapshot());"
        )
    else:
        query = connection.execute(
            "SET TRANSACTION ISOLATION LEVEL SERIALIZABLE, READ ONLY, DEFERRABLE;"
            "SELECT txid_snapshot_xmin(txid_current_snapshot());"
        )
    # DEFERRABLE prevents query cancelling due to conflicts but requires SERIALIZABLE mode
    # which is not available in recovery.
    xmin = query.scalar()  # lowest xid that is still in progress
    return xmin


class Indexer(object):
    def __init__(self, registry):
        self.es = registry[ELASTIC_SEARCH]
        self.esstorage = registry[STORAGE]
        self.index = registry.settings['snovault.elasticsearch.index']
        self.queue_server = None
        self.queue_server_backup = None
        self.queue_worker = None
        self.chunk_size = None
        self.batch_size = None
        self.worker_runs = []
        if registry.settings.get('indexer'):
            self._setup_queues(registry)

    def check_log_indexing_times(self, update_infos):
        if self.indexer_initial_log and not os.path.exists(self.indexer_initial_log_path):
            log.warning('Logging indexing data to %s', self.indexer_initial_log_path)
            counter = 0
            with open(self.indexer_initial_log_path, 'w', encoding='utf-8') as file_handler:
                for update_info in update_infos:
                    str_update_info = json.dumps(update_info, ensure_ascii=False)
                    file_handler.write(str_update_info + '\n')
                    counter += 1
            log.warning('Logged %d uuids.  One per line' % counter)

    def _setup_queues(self, registry):
        '''Init helper - Setup server and worker queues'''
        self.index = registry.settings['snovault.elasticsearch.index']
        self.indexer_initial_log = asbool(registry.settings.get('indexer_initial_log', False))
        self.indexer_initial_log_path = registry.settings.get('indexer_initial_log_path')
        try:
            self.indexer_short_uuids = int(registry.settings.get('indexer_short_uuids'))
        except Exception as ecp:
            log.warning('indexer_short_uuids could not be cast to int.  Defaulting to all.')
            self.indexer_short_uuids = 0
        queue_type = registry.settings.get('queue_type', None)
        is_queue_server = asbool(registry.settings.get('queue_server'))
        is_queue_worker = asbool(registry.settings.get('queue_worker'))
        queue_options = self._get_queue_options(registry)
        self.chunk_size = queue_options['chunk_size']
        self.batch_size = queue_options['batch_size']
        if is_queue_server:
            cp_q_ops = queue_options.copy()
            cp_q_ops['batch_size'] = cp_q_ops['get_size']
            self.queue_server_backup = SimpleUuidServer(cp_q_ops)
            if (
                    not queue_type or
                    queue_type == DEFAULT_QUEUE or
                    queue_type not in registry['available_queues']
                ):
                self.queue_type = DEFAULT_QUEUE
                self.queue_server = self.queue_server_backup
                self.queue_server_backup = None
            elif 'UuidQueue' in registry:
                try:
                    queue_options['uuid_len'] = 36
                    self.queue_server = registry['UuidQueue'](
                        queue_options['queue_name'],
                        queue_type,
                        queue_options,
                    )
                except Exception as exp:  # pylint: disable=broad-except
                    log.warning(repr(exp))
                    log.warning(
                        'Failed to initialize UuidQueue. Switching to backup.'
                    )
                    self._serve_object_switch_queue()
                else:
                    self.queue_type = queue_type
            else:
                log.error('No queue available for Indexer')
            if self.queue_server and is_queue_worker:
                self.queue_worker = self.queue_server.get_worker()
            log.warning('Primary indexer queue type: %s', self.queue_type)

    @staticmethod
    def _get_queue_options(registry):
        '''Init helper - Extract queue options from registry settings'''
        queue_name = registry.settings.get('queue_name', 'indxQ')
        queue_worker_processes = int(
            registry.settings.get('queue_worker_processes', 1)
        )
        queue_worker_chunk_size = int(
            registry.settings.get('queue_worker_chunk_size', 1024)
        )
        queue_worker_batch_size = int(
            registry.settings.get('queue_worker_batch_size', 5000)
        )
        queue_worker_get_size = int(
            registry.settings.get('queue_worker_get_size', 2000000)
        )
        # Only Used for Redis Queues
        queue_host = registry.settings.get('queue_host', 'localhost')
        queue_port = registry.settings.get('queue_port', 6379)
        queue_db = registry.settings.get('queue_db', 2)
        return {
            'queue_name': queue_name,
            'processes': queue_worker_processes,
            'chunk_size': queue_worker_chunk_size,
            'batch_size': queue_worker_batch_size,
            'get_size': queue_worker_get_size,
            'host': queue_host,
            'port': queue_port,
            'db': queue_db,
        }

    def _serve_object_switch_queue(self, set_worker=False):
        # If a non simple queue server fails we end up back here
        # on the next pass of the index listenter with the original
        # list of uuids(given the functionality of the indexer_state)
        # For any failure we switch to simple backup server permanently.
        self.queue_server = self.queue_server_backup
        self.queue_server_backup = None
        self.queue_type = DEFAULT_QUEUE
        if set_worker:
            self.queue_worker = self.queue_server.get_worker()
    

    def _serve_objects_init(self, uuids):
        err_msg = 'Cannot initialize indexing process: '
        try:
            is_indexing = self.queue_server.is_indexing()
            if is_indexing:
                return err_msg + 'Already Indexing'
            elif not uuids:
                return err_msg + 'No uuids given to Indexer.serve_objects'
        except Exception as exp:  # pylint: disable=broad-except
            log.warning(repr(exp))
            if self.queue_server_backup:
                log.warning('uuid init issue:  Switching to simple server.')
                self._serve_object_switch_queue(set_worker=True)
            else:
                return err_msg + 'Cannot failover to simple queue'
        return None

    def _serve_objects_load_uuids(self, uuids):
        err_msg = None
        try:
            uuids_loaded_len = self.queue_server.load_uuids(uuids)
            if not uuids_loaded_len:
                err_msg = 'Uuids given to Indexer.serve_objects failed to load'
            elif uuids_loaded_len != len(uuids):
                err_msg = (
                    'Uuids given to Indexer.serve_objects '
                    'failed to all load. {} of {} only'.format(
                        uuids_loaded_len,
                        len(uuids),
                    )
                )
        except Exception as exp:  # pylint: disable=broad-except
            log.warning(repr(exp))
            err_msg = 'Indexer load uuids failed.'
            if self.queue_server_backup:
                log.warning('uuid load issue:  Switching to simple server')
                self._serve_object_switch_queue(set_worker=True)
        return err_msg

    def serve_objects(
            self,
            request,
            uuids,
            xmin,
            snapshot_id=None,
            restart=False,
            timeout=None,
        ):
        '''Run indexing process with queue server and optional worker'''
        # pylint: disable=too-many-arguments
        errors = []
        err_msg = self._serve_objects_init(uuids)
        if err_msg:
            return None, errors, err_msg
        # Check for shorting uuids
        if self.indexer_short_uuids:
            short_uuids = []
            for count, uuid in enumerate(uuids, 1):
                if count > self.indexer_short_uuids:
                    break
                short_uuids.append(uuid)
            log.warning(
                'Shorting %d uuids to %d.  New list is %d uuids.',
                len(uuids),
                self.indexer_short_uuids,
                len(short_uuids)
            )
            uuids = short_uuids
        err_msg = self._serve_objects_load_uuids(uuids)
        if err_msg:
            return None, errors, err_msg
        # Run Process Loop
        start_time = time.time()
        self.worker_runs = []
        update_infos = []
        while self.queue_server.is_indexing(errs_cnt=len(errors)):
            if self.queue_worker and not self.queue_worker.is_running:
                # Server Worker
                update_infos, uuids_ran = self.run_worker(
                    request, xmin, snapshot_id, restart
                )
                if not uuids_ran:
                    break
                self.worker_runs.append({
                    'worker_id':self.queue_worker.worker_id,
                    'uuids': uuids_ran,
                })
            # Handling Errors must happen or queue will not stop
            batch_errors = self.queue_server.pop_errors()
            for error in batch_errors:
                errors.append(error)
            if timeout and time.time() - start_time > timeout:
                err_msg = 'Indexer sleep timeout'
                break
        self.queue_server.close_indexing()
        return update_infos, errors, err_msg

    def run_worker(self, request, xmin, snapshot_id, restart):
        '''Run the uuid queue worker'''
        batch_uuids = self.queue_worker.get_uuids(get_all=False)
        log.warning(
            'running %s with %d',
            self.queue_worker.worker_id,
            len(batch_uuids),
        )
        update_infos = []
        if batch_uuids:
            self.queue_worker.is_running = True
            batch_update_infos, batch_errors = self.update_objects(
                request,
                batch_uuids,
                xmin,
                snapshot_id=snapshot_id,
                restart=restart,
            )
            update_infos.extend(batch_update_infos)
            batch_results = {
                'errors': batch_errors,
                'successes': len(batch_uuids) - len(batch_errors),
            }
            err_msg = self.queue_worker.update_finished(batch_results)
            if err_msg:
                log.warning('Issue closing worker: %s', err_msg)
            self.queue_worker.is_running = False
            return update_infos, len(batch_uuids)
        else:
            log.warning('No uudis to run %d', self.queue_worker.get_cnt)
        return update_infos, None

    def update_objects(
            self,
            request,
            uuids,
            xmin,
            snapshot_id=None,
            restart=False,
        ):
        # pylint: disable=too-many-arguments, unused-argument
        '''Run indexing process on uuids'''
        errors = []
        update_infos = []
        for i, uuid in enumerate(uuids):
            update_info = self.update_object(self.es, request, uuid, xmin)
            update_info['return_time'] = time.time()
            update_infos.append(update_info)
            error = update_info.get('error')
            if error is not None:
                print('Error', error)
                errors.append(error)
            if (i + 1) % 1000 == 0:
                log.info('Indexing %d', i + 1)
        return update_infos, errors

    @staticmethod
    def update_object(encoded_es, request, uuid, xmin, restart=False):
        update_info = {
            'uuid': uuid,
            'xmin': xmin,
            'start_time': time.time(),
            'end_time': None,
            'run_time': None,
            'error': None,
            'return_time': None,
        }
        req_info = {
            'start_time': None,
            'end_time': None,
            'run_time': None,
            'errors': [],
            'url': None
        }
        es_info = {
            'start_time': None,
            'end_time': None,
            'run_time': None,
            'backoffs': {},
            'item_type': None,
        }
        request.datastore = 'database'
        last_exc = None
        req_info['start_time'] = time.time()
        backoff = 0
        try:
            req_info['url'] ='/%s/@@index-data/' % uuid
            doc = request.embed(req_info['url'], as_user='INDEXER')
        except StatementError:
            # Can't reconnect until invalid transaction is rolled back
            raise
        except Exception as e:
            msg = 'Error rendering /%s/@@index-data' % uuid
            log.error(msg, exc_info=True)
            last_exc = repr(e)
            req_info['errors'].append(
                {
                    'backoff': backoff,
                    'msg': msg,
                    'last_exc': last_exc,
                }
            )
        req_info['end_time'] = time.time()
        req_info['run_time'] = req_info['end_time'] - req_info['start_time']
        if last_exc is None:
            es_info['start_time'] = time.time()
            es_info['item_type'] = doc['item_type']
            do_break = False
            for backoff in [0, 10, 20, 40, 80]:
                time.sleep(backoff)
                backoff_info = {
                    'start_time': time.time(),
                    'end_time': None,
                    'run_time': None,
                    'error': None,
                }
                try:
                    encoded_es.index(
                        index=doc['item_type'], doc_type=doc['item_type'], body=doc,
                        id=str(uuid), version=xmin, version_type='external_gte',
                        request_timeout=30,
                    )
                except StatementError:
                    # Can't reconnect until invalid transaction is rolled back
                    raise
                except ConflictError:
                    msg = 'Conflict indexing %s at version %d' % (uuid, xmin)
                    log.warning(msg)
                    backoff_info['error'] = {
                        'msg': msg,
                        'last_exc': None,
                    }
                    do_break = True
                except (ConnectionError, ReadTimeoutError, TransportError) as e:
                    msg = 'Retryable error indexing %s: %r' % (uuid, e)
                    log.warning(msg)
                    last_exc = repr(e)
                    backoff_info['error'] = {
                        'msg': msg,
                        'last_exc': last_exc,
                    }
                except Exception as e:
                    msg = 'Error indexing %s' % (uuid)
                    log.error(msg, exc_info=True)
                    last_exc = repr(e)
                    backoff_info['error'] = {
                        'msg': msg,
                        'last_exc': None,
                    }
                    do_break = True
                else:
                    # Get here on success and outside of try
                    do_break = True
                end_time = time.time()
                backoff_info['end_time'] = end_time
                backoff_info['run_time'] = end_time - backoff_info['start_time']
                es_info['backoffs'][str(backoff)] = backoff_info
                if do_break:
                    break
            es_info['end_time'] = time.time()
            es_info['run_time'] = es_info['end_time'] - es_info['start_time']
        update_info['req_info'] = req_info
        update_info['es_info'] = es_info
        if last_exc:
            update_info['error'] = {
                'error_message': last_exc,
                'timestamp': datetime.datetime.now().isoformat(),
                'uuid': str(uuid)
            }
        end_time = time.time()
        update_info['end_time'] = end_time
        update_info['run_time'] = end_time - update_info['start_time']
        return update_info

    def shutdown(self):
        pass



# -*- coding: utf-8 -*-
import uuid
from pyramid.httpexceptions import HTTPUnprocessableEntity
from pyramid.view import (
    view_config,
)
from pyramid.security import (
    Allow,
    Deny,
    Everyone,
)
from .base import (
    Item,
)
from snovault import (
    CONNECTION,
    DBSESSION,
    calculated_property,
    collection,
    load_schema,
)
from snovault.calculated import calculate_properties
from snovault.resource_views import item_view_object
from snovault.util import expand_path
from snovault.crud_views import collection_add
from snovault.schema_utils import validate_request
from snovault.storage import User as AuthUser


ONLY_ADMIN_VIEW_DETAILS = [
    (Allow, 'group.admin', ['view', 'view_details', 'edit']),
    (Allow, 'group.read-only-admin', ['view', 'view_details']),
    (Allow, 'remoteuser.INDEXER', ['view']),
    (Allow, 'remoteuser.EMBED', ['view']),
    (Deny, Everyone, ['view', 'view_details', 'edit']),
]

USER_ALLOW_CURRENT = [
    (Allow, Everyone, 'view'),
] + ONLY_ADMIN_VIEW_DETAILS

USER_DELETED = [
    (Deny, Everyone, 'visible_for_edit')
] + ONLY_ADMIN_VIEW_DETAILS


@collection(
    name='users',
    unique_key='user:email',
    properties={
        'title': 'Snowflake  Users',
        'description': 'Listing of current Snowflake users',
    },
    acl=[])
class User(Item):
    item_type = 'user'
    schema = load_schema('snowflakes:schemas/user.json')
    # Avoid access_keys reverse link so editing access keys does not reindex content.
    embedded = [
        'lab',
    ]
    STATUS_ACL = {
        'current': [(Allow, 'role.owner', ['edit', 'view_details'])] + USER_ALLOW_CURRENT,
        'deleted': USER_DELETED,
        'replaced': USER_DELETED,
        'disabled': ONLY_ADMIN_VIEW_DETAILS,
    }

    @calculated_property(schema={
        "title": "Title",
        "type": "string",
    })
    def title(self, first_name, last_name):
        return u'{} {}'.format(first_name, last_name)

    def __ac_local_roles__(self):
        owner = 'userid.%s' % self.uuid
        return {owner: 'role.owner'}

    def _update(self, properties, sheets=None):
        '''
        overwritting this so as to create an associated 'web user'
        for each user object, if it doesn't have on already
        '''
        super(User, self)._update(properties, sheets)

        # after we save if we don't have AuthUser create it
        db = self.registry[DBSESSION]
        email = properties['email']
        if db.query(AuthUser).filter_by(email=email).first():
            return

        name ="%s %s" % (properties['first_name'], properties['last_name'])
        # we don't keep passwords in our user schema, so just create a random one here
        pwd = str(uuid.uuid4())

        # this is executed in an existing session, so it should get saved eventually
        auth_user = AuthUser(email, pwd, name)
        db.add(auth_user)

    @calculated_property(schema={
        "title": "Access Keys",
        "type": "array",
        "items": {
            "type": ['string', 'object'],
            "linkFrom": "AccessKey.user",
        },
    }, category='page')
    def access_keys(self, request):
        if not request.has_permission('view_details'):
            return
        uuids = self.registry[CONNECTION].get_rev_links(self.model, 'user', 'AccessKey')
        objects = (request.embed('/', str(uuid), '@@object') for uuid in uuids)
        return [obj for obj in objects if obj['status'] not in ('deleted', 'replaced')]


@view_config(context=User, permission='view', request_method='GET', name='page')
def user_page_view(context, request):
    if request.has_permission('view_details'):
        properties = item_view_object(context, request)
    else:
        item_path = request.resource_path(context)
        properties = request.embed(item_path, '@@object')
    for path in context.embedded:
        expand_path(request, properties, path)
    calculated = calculate_properties(context, request, properties, category='page')
    properties.update(calculated)
    return properties


@view_config(context=User, permission='view', request_method='GET',
             name='object')
def user_basic_view(context, request):
    properties = item_view_object(context, request)
    filtered = {}
    for key in ['@id', '@type', 'uuid', 'lab', 'title']:
        try:
            filtered[key] = properties[key]
        except KeyError:
            pass
    return filtered


@view_config(context=User.Collection, permission='add', request_method='POST',
             physical_path="/users")
def user_add(context,request):
    ''' if we have a password in our request, create and auth entry
     for the user as well
     '''

    #do we have valid data
    pwd = request.json.get('password', None)
    pwd_less_data = request.json.copy()

    if pwd is not None:
        del pwd_less_data['password']

    validate_request(context.type_info.schema, request, pwd_less_data)

    if request.errors:
        return HTTPUnprocessableEntity(json={'errors':request.errors},
                                     content_type='application/json')

    # this will create an AuthUser with random password
    result = collection_add(context, request)
    if result:
        email = request.json.get('email')
        pwd = request.json.get('password', None)

        if pwd is not None:
            # now update the password
            db = request.registry['dbsession']
            auth_user = db.query(AuthUser).filter_by(email=email).first()
            auth_user.password = pwd

            import transaction
            transaction.commit()

    return result


@calculated_property(context=User, category='user_action')
def impersonate(request):
    # This is assuming the user_action calculated properties
    # will only be fetched from the current_user view,
    # which ensures that the user represented by 'context' is also an effective principal
    if request.has_permission('impersonate'):
        return {
            'id': 'impersonate',
            'title': 'Impersonate User…',
            'href': '/#!impersonate-user',
        }


@calculated_property(context=User, category='user_action')
def profile(context, request):
    return {
        'id': 'profile',
        'title': 'Profile',
        'href': request.resource_path(context),
    }


@calculated_property(context=User, category='user_action')
def signout(context, request):
    return {
        'id': 'signout',
        'title': 'Sign out',
        'trigger': 'logout',
    }



import base64
import codecs
import copy
import json
import os

from pyramid.path import (
    AssetResolver,
    caller_package,
)
from pyramid.session import SignedCookieSessionFactory
from sqlalchemy import engine_from_config
from webob.cookies import JSONSerializer
from .json_renderer import json_renderer
from pyramid.settings import (
    asbool,
)

STATIC_MAX_AGE = 0


def json_asset(spec, **kw):
    utf8 = codecs.getreader("utf-8")
    asset = AssetResolver(caller_package()).resolve(spec)
    return json.load(utf8(asset.stream()), **kw)


def static_resources(config):
    from pkg_resources import resource_filename
    import mimetypes
    mimetypes.init()
    mimetypes.init([resource_filename('snowflakes', 'static/mime.types')])
    config.add_static_view('static', 'static', cache_max_age=STATIC_MAX_AGE)
    config.add_static_view('profiles', 'schemas', cache_max_age=STATIC_MAX_AGE)

    favicon_path = '/static/img/favicon.ico'
    if config.route_prefix:
        favicon_path = '/%s%s' % (config.route_prefix, favicon_path)
    config.add_route('favicon.ico', 'favicon.ico')

    def favicon(request):
        subreq = request.copy()
        subreq.path_info = favicon_path
        response = request.invoke_subrequest(subreq)
        return response

    config.add_view(favicon, route_name='favicon.ico')


def changelogs(config):
    config.add_static_view(
        'profiles/changelogs', 'schemas/changelogs', cache_max_age=STATIC_MAX_AGE)


def configure_engine(settings):
    settings = copy.deepcopy(settings)
    engine_url = os.environ.get("SQLALCHEMY_URL") or settings['sqlalchemy.url']
    settings["sqlalchemy.url"] = engine_url
    engine_opts = {}
    if engine_url.startswith('postgresql'):
        if settings.get('indexer_worker'):
            application_name = 'indexer_worker'
        elif settings.get('indexer'):
            application_name = 'indexer'
        else:
            application_name = 'app'
        engine_opts = dict(
            isolation_level='REPEATABLE READ',
            json_serializer=json_renderer.dumps,
            connect_args={'application_name': application_name}
        )
    engine = engine_from_config(settings, 'sqlalchemy.', **engine_opts)
    if engine.url.drivername == 'postgresql':
        timeout = settings.get('postgresql.statement_timeout')
        if timeout:
            timeout = int(timeout) * 1000
            set_postgresql_statement_timeout(engine, timeout)
    return engine


def set_postgresql_statement_timeout(engine, timeout=20 * 1000):
    """ Prevent Postgres waiting indefinitely for a lock.
    """
    from sqlalchemy import event
    import psycopg2

    @event.listens_for(engine, 'connect')
    def connect(dbapi_connection, connection_record):
        cursor = dbapi_connection.cursor()
        try:
            cursor.execute("SET statement_timeout TO %d" % timeout)
        except psycopg2.Error:
            dbapi_connection.rollback()
        finally:
            cursor.close()
            dbapi_connection.commit()


def json_from_path(path, default=None):
    if path is None:
        return default
    return json.load(open(path))


def configure_dbsession(config):
    from snovault import DBSESSION
    settings = config.registry.settings
    DBSession = settings.pop(DBSESSION, None)
    if DBSession is None:
        engine = configure_engine(settings)

        if asbool(settings.get('create_tables', False)):
            from snovault.storage import Base
            Base.metadata.create_all(engine)

        import snovault.storage
        import zope.sqlalchemy
        from sqlalchemy import orm

        DBSession = orm.scoped_session(orm.sessionmaker(bind=engine))
        zope.sqlalchemy.register(DBSession)
        snovault.storage.register(DBSession)

    config.registry[DBSESSION] = DBSession


def session(config):
    """ To create a session secret on the server:

    $ cat /dev/urandom | head -c 256 | base64 > session-secret.b64
    """
    settings = config.registry.settings
    if 'session.secret' in settings:
        secret = settings['session.secret'].strip()
        if secret.startswith('/'):
            secret = open(secret).read()
            secret = base64.b64decode(secret)
    else:
        secret = os.urandom(256)
    # auth_tkt has no timeout set
    # cookie will still expire at browser close
    if 'session.timeout' in settings:
        timeout = int(settings['session.timeout'])
    else:
        timeout = 60 * 60 * 24
    session_factory = SignedCookieSessionFactory(
        secret=secret,
        timeout=timeout,
        reissue_time=2**32,  # None does not work
        serializer=JSONSerializer(),
    )
    config.set_session_factory(session_factory)


def app_version(config):
    import hashlib
    import os
    import subprocess
    try:
        version = subprocess.check_output(
            ['git', '-C', os.path.dirname(__file__), 'describe']).decode('utf-8').strip()
    except:
        version = "travis_version"
    diff = subprocess.check_output(
        ['git', '-C', os.path.dirname(__file__), 'diff', '--no-ext-diff'])
    if diff:
        version += '-patch' + hashlib.sha1(diff).hexdigest()[:7]
    config.registry.settings['snovault.app_version'] = version


def main(global_config, **local_config):
    """ This function returns a Pyramid WSGI application.
    """
    settings = global_config
    settings.update(local_config)

    settings['snovault.jsonld.namespaces'] = json_asset('encoded:schemas/namespaces.json')
    settings['snovault.jsonld.terms_namespace'] = 'https://www.encodeproject.org/terms/'
    settings['snovault.jsonld.terms_prefix'] = 'encode'
    settings['snovault.elasticsearch.index'] = 'encoded'
    hostname_command = settings.get('hostname_command', '').strip()
    if hostname_command:
        hostname = subprocess.check_output(hostname_command, shell=True).strip()
        settings.setdefault('persona.audiences', '')
        settings['persona.audiences'] += '\nhttp://%s' % hostname
        settings['persona.audiences'] += '\nhttp://%s:6543' % hostname

    config = Configurator(settings=settings)
    from snovault.elasticsearch import APP_FACTORY
    config.registry[APP_FACTORY] = main  # used by mp_indexer
    config.include(app_version)

    config.include('pyramid_multiauth')  # must be before calling set_authorization_policy
    from pyramid_localroles import LocalRolesAuthorizationPolicy
    # Override default authz policy set by pyramid_multiauth
    config.set_authorization_policy(LocalRolesAuthorizationPolicy())
    config.include(session)
    config.include('.persona')

    config.include(configure_dbsession)
    config.include('snovault')
    config.commit()  # commit so search can override listing

    # Render an HTML page to browsers and a JSON document for API clients
    config.include('.renderers')
    config.include('.authentication')
    config.include('.server_defaults')
    config.include('.types')
    config.include('.root')
    config.include('.batch_download')
    config.include('.visualization')

    if 'elasticsearch.server' in config.registry.settings:
        config.include('snovault.elasticsearch')
        config.include('.search')

    if 'snp_search.server' in config.registry.settings:
        addresses = aslist(config.registry.settings['snp_search.server'])
        config.registry['snp_search'] = Elasticsearch(
            addresses,
            serializer=PyramidJSONSerializer(json_renderer),
            connection_class=TimedUrllib3HttpConnection,
            retry_on_timeout=True,
            timeout=60,
            maxsize=50
        )
        config.include('.region_search')
        config.include('.peak_indexer')
    config.include(static_resources)
    config.include(changelogs)

    config.registry['ontology'] = json_from_path(settings.get('ontology_path'), {})

    if asbool(settings.get('testing', False)):
        config.include('.tests.testing_views')

    # Load upgrades last so that all views (including testing views) are
    # registered.
    config.include('.upgrade')
    config.include('.audit')

    app = config.make_wsgi_app()

    workbook_filename = settings.get('load_workbook', '')
    load_test_only = asbool(settings.get('load_test_only', False))
    docsdir = settings.get('load_docsdir', None)
    if docsdir is not None:
        docsdir = [path.strip() for path in docsdir.strip().split('\n')]
    if workbook_filename:
        load_workbook(app, workbook_filename, docsdir, test=load_test_only)

    return app



import pytest
pytestmark = pytest.mark.storage


def test_storage_creation(session):
    from snovault.storage import (
        PropertySheet,
        CurrentPropertySheet,
        TransactionRecord,
        Blob,
        Key,
        Link,
    )
    assert session.query(PropertySheet).count() == 0
    assert session.query(CurrentPropertySheet).count() == 0
    assert session.query(TransactionRecord).count() == 0
    assert session.query(Blob).count() == 0
    assert session.query(Key).count() == 0
    assert session.query(Link).count() == 0


def test_transaction_record(session):
    from snovault.storage import (
        Resource,
        PropertySheet,
        TransactionRecord,
    )
    name = 'testdata'
    props1 = {'foo': 'bar'}
    resource = Resource('test_item')
    session.add(resource)
    propsheet = PropertySheet(name=name, properties=props1, rid=resource.rid)
    session.add(propsheet)
    session.flush()
    assert session.query(PropertySheet).count() == 1
    propsheet = session.query(PropertySheet).one()
    assert session.query(TransactionRecord).count() == 1
    record = session.query(TransactionRecord).one()
    assert record.tid
    assert propsheet.tid == record.tid


def test_transaction_record_rollback(session):
    import transaction
    import uuid
    from snovault.storage import Resource
    rid = uuid.uuid4()
    resource = Resource('test_item', {'': {}}, rid=rid)
    session.add(resource)
    transaction.commit()
    transaction.begin()
    sp = session.begin_nested()
    resource = Resource('test_item', {'': {}}, rid=rid)
    session.add(resource)
    with pytest.raises(Exception):
        sp.commit()
    sp.rollback()
    resource = Resource('test_item', {'': {}})
    session.add(resource)
    transaction.commit()


def test_current_propsheet(session):
    from snovault.storage import (
        CurrentPropertySheet,
        Resource,
        PropertySheet,
        TransactionRecord,
    )
    name = 'testdata'
    props1 = {'foo': 'bar'}
    resource = Resource('test_item', {name: props1})
    session.add(resource)
    session.flush()
    resource = session.query(Resource).one()
    assert resource.rid
    assert resource[name] == props1
    propsheet = session.query(PropertySheet).one()
    assert propsheet.sid
    assert propsheet.rid == resource.rid
    current = session.query(CurrentPropertySheet).one()
    assert current.sid == propsheet.sid
    assert current.rid == resource.rid
    record = session.query(TransactionRecord).one()
    assert record.tid
    assert propsheet.tid == record.tid


def test_current_propsheet_update(session):
    from snovault.storage import (
        CurrentPropertySheet,
        Resource,
        PropertySheet,
    )
    name = 'testdata'
    props1 = {'foo': 'bar'}
    resource = Resource('test_item', {name: props1})
    session.add(resource)
    session.flush()
    resource = session.query(Resource).one()
    props2 = {'foo': 'baz'}
    resource[name] = props2
    session.flush()
    resource = session.query(Resource).one()
    session.flush()
    assert resource[name] == props2
    assert session.query(PropertySheet).count() == 2
    assert [propsheet.properties for propsheet in resource.data[name].history] == [props1, props2]
    current = session.query(CurrentPropertySheet).one()
    assert current.sid


def test_delete_simple(session, storage):
    from snovault.storage import (
        Resource,
        Key,
        PropertySheet,
        CurrentPropertySheet,
        TransactionRecord,
    )
    name = 'testdata'
    props1 = {'foo': 'bar'}
    resource = Resource('test_item', {name: props1})
    session.add(resource)
    session.flush()
    resource = session.query(Resource).one()
    check = storage.get_by_uuid(str(resource.rid))
    assert check[name] == props1
    # add a key
    testname = 'foo'
    key = Key(rid=resource.rid, name=testname, value=props1[testname])
    session.add(key)
    session.flush()
    assert session.query(Key).count() == 1

    propsheet = session.query(PropertySheet).one()
    assert propsheet.sid
    assert propsheet.rid == resource.rid
    current = session.query(CurrentPropertySheet).one()
    assert current.sid == propsheet.sid
    assert current.rid == resource.rid

    storage.delete_by_uuid(str(resource.rid))
    check_post = storage.get_by_uuid(str(resource.rid))
    assert not check_post
    assert session.query(Key).count() == 0
    assert session.query(PropertySheet).count() == 0
    assert session.query(CurrentPropertySheet).count() == 0


def test_delete_compound(session, storage):
    from snovault.storage import (
        CurrentPropertySheet,
        Resource,
        PropertySheet,
        Key,
    )
    name = 'testdata'
    props1 = {'foo': 'bar'}
    resource = Resource('test_item', {name: props1})
    session.add(resource)
    session.flush()
    resource = session.query(Resource).one()
    check = storage.get_by_uuid(str(resource.rid))
    assert check[name] == props1
    # add a key
    testname = 'foo'
    key = Key(rid=resource.rid, name=testname, value=props1[testname])
    session.add(key)
    session.flush()
    assert session.query(Key).count() == 1

    props2 = {'foo': 'baz'}
    resource[name] = props2
    session.flush()
    resource = session.query(Resource).one()
    session.flush()
    assert resource[name] == props2
    assert session.query(PropertySheet).count() == 2
    assert [propsheet.properties for propsheet in resource.data[name].history] == [props1, props2]
    current = session.query(CurrentPropertySheet).one()
    assert current.sid

    storage.delete_by_uuid(str(resource.rid))
    check_post = storage.get_by_uuid(str(resource.rid))
    assert not check_post
    assert session.query(Key).count() == 0
    assert session.query(PropertySheet).count() == 0
    assert session.query(CurrentPropertySheet).count() == 0


def test_keys(session):
    from sqlalchemy.orm.exc import FlushError
    from snovault.storage import (
        Resource,
        Key,
    )
    name = 'testdata'
    props1 = {'foo': 'bar'}
    resource = Resource('test_item', {name: props1})
    session.add(resource)
    session.flush()
    resource = session.query(Resource).one()

    testname = 'foo'
    key = Key(rid=resource.rid, name=testname, value=props1[testname])
    session.add(key)
    session.flush()
    assert session.query(Key).count() == 1
    othertest = 'foofoo'
    othervalue = 'barbar'
    key2 = Key(rid=resource.rid, name=othertest, value=othervalue)
    session.add(key2)
    session.flush()
    assert session.query(Key).count() == 2
    props2 = {'foofoo': 'barbar'}
    resource2 = Resource('test_item', {name: props2})
    session.add(resource2)
    session.flush()
    key3 = Key(rid=resource2.rid, name=testname, value=props1[testname])
    session.add(key3)
    with pytest.raises(FlushError):
        session.flush()


def test_S3BlobStorage_boto3(mocker):
    from snovault.storage import S3BlobStorage
    mocker.patch('boto3.Session.resource')
    mocker.patch('boto3.Session.client')
    bucket = 'test'
    mock_key = mocker.Mock()
    storage = S3BlobStorage(bucket)
    storage.bucket.name = bucket
    storage.store_conn.return_value = mock_key
    download_meta = {'blob_id': '123'}
    storage.store_blob('data', download_meta)
    assert download_meta['bucket'] == 'test'
    assert 'key' in download_meta
    assert storage.store_conn.put_object.call_count == 1
    storage.read_conn.return_value = mock_key
    storage.read_conn.get_object()['Body'].read.return_value = 'data'
    data = storage.get_blob(download_meta)
    assert data == 'data'
    assert storage.read_conn.get_object.call_count == 2


def test_S3BlobStorage_boto3_with_header(mocker):
    from snovault.storage import S3BlobStorage
    mocker.patch('boto3.Session.resource')
    mocker.patch('boto3.Session.client')
    bucket = 'test'
    mock_key = mocker.Mock()
    storage = S3BlobStorage(bucket)
    storage.bucket.name = bucket
    storage.store_conn.return_value = mock_key
    download_meta = {'blob_id': '123', 'type': 'text/plain'}
    storage.store_blob('data', download_meta)
    assert 'type' in download_meta
    assert download_meta['type'] == 'text/plain'


def test_S3BlobStorage_boto3_get_blob_url_for_s3_file(mocker):
    from snovault.storage import S3BlobStorage
    mocker.patch('boto3.Session.resource')
    mocker.patch('boto3.Session.client')
    bucket = 'test'
    storage = S3BlobStorage(bucket)
    storage.bucket.name = bucket
    download_meta = {'blob_id': '123'}
    storage.read_conn.generate_presigned_url.return_value = 'http://testurl'
    url = storage.get_blob_url(download_meta)
    assert url == 'http://testurl'
    storage.read_conn.generate_presigned_url.assert_called_once_with(
        Params={'Key': '123', 'Bucket': 'test'}, ExpiresIn=129600, ClientMethod='get_object'
    )



import pytest

from redis import StrictRedis

from snovault.local_storage import LocalStoreClient


def _get_client(local_settings):
    return LocalStoreClient(
        db_index=local_settings['local_storage_redis_index'],
        host=local_settings['local_storage_host'],
        port=local_settings['local_storage_port'],
        socket_timeout=local_settings['local_storage_timeout'],
    )


def test_local_storage_server_fixture(app_settings):
    local_store = _get_client(app_settings)
    try:
        local_store.ping()
    except Exception as excp:  # pylint: disable=broad-except
        print(excp)
        assert False
    assert True


class TestLocalStore():
    dict_key = 'fakedictkey'
    item_key = 'fakeitemkey'
    list_key = 'fakelistkey'

    @pytest.fixture(autouse=True)
    def setup_method(self, app_settings):
        '''
        Add local store to test class and cleans up standard redis keys
        - Uses the exposed redis client directly
        '''
        self.local_store = _get_client(app_settings)
        self.local_store.client.delete(self.dict_key)
        self.local_store.client.delete(self.item_key)
        for item in self.local_store.client.lrange(self.list_key, 0, -1):
            self.local_store.client.delete(item)
        self.local_store.client.delete(self.list_key)

    def test_init(self):
        assert hasattr(self.local_store, 'local_tz')
        assert isinstance(self.local_store.local_tz, str)
        assert self.local_store.local_tz == 'GMT'
        assert hasattr(self.local_store, 'client')
        assert isinstance(self.local_store.client, StrictRedis)

    def test_get_tag(self):
        config_tag = 'localtesting'
        tag = self.local_store.get_tag(config_tag)
        assert isinstance(tag, str)
        assert len(tag) == len(config_tag) + len(':') + 4
        tag = self.local_store.get_tag(config_tag, num_bytes=4)
        assert len(tag) == len(config_tag) + len(':') + 8

    def test_ping(self):
        try:
            self.local_store.ping()
        except Exception as excp:  # pylint: disable=broad-except
            assert False
        assert True

    def test_dict_get(self):
        hash_dict = self.local_store.dict_get(self.dict_key)
        assert isinstance(hash_dict, dict)
        assert hash_dict == {}

    def test_dict_set(self):
        update_dict = {'somekey': 'somevalue'}
        self.local_store.dict_set(self.dict_key, update_dict)
        hash_dict = self.local_store.dict_get(self.dict_key)
        assert hash_dict == update_dict

    def test_item_get(self):
        item = self.local_store.item_get(self.item_key)
        assert item is None

    def test_item_set(self):
        set_item = 'someitem'
        self.local_store.item_set(self.item_key, set_item)
        item = self.local_store.item_get(self.item_key)
        assert isinstance(item, str)
        assert item == set_item

    def test_list_add(self):
        # Lists are LIFO
        items_to_add = ['A', 'B', 'C']
        for item in items_to_add:
            self.local_store.list_add(self.list_key, item)
        items = self.local_store.list_get(self.list_key, 0, -1)
        items.sort()
        assert items_to_add == items

    def test_list_get(self):
        # Lists are LIFO
        items_to_add = ['A', 'B', 'C', 'D', 'E']
        for item in items_to_add:
            self.local_store.list_add(self.list_key, item)
        items = self.local_store.list_get(self.list_key, 0, 0)
        assert items[0] == items_to_add[-1]
        items = self.local_store.list_get(self.list_key, 1, 3)
        items_to_add.sort(reverse=True)
        assert items == items_to_add[1:4]



import pytest


def test_heterogeneous_stream():
    from snovault.elasticsearch.indexer_state import heterogeneous_stream
    gm = {'e': (x for x in [1, 2, 3, 4, 5])}
    assert list(heterogeneous_stream(gm)) == [1, 2, 3, 4, 5]
    gm = {
        'e': (x for x in [1, 2, 3, 4, 5]),
        'f': (x for x in ['a', 'b', 'c'])
    }
    assert list(heterogeneous_stream(gm)) == [1, 'a', 2, 'b', 3, 'c', 4, 5]
    gm = {
        'e': (x for x in [1, 2, 3, 4, 5]),
        'f': (x for x in ['a', 'b', 'c']),
        'g': (x for x in (t for t in (None, True, False, None, None, True)))
    }
    assert list(heterogeneous_stream(gm)) == [
        1, 'a', 2, 'b', True, 3, 'c', False, 4, 5, True
    ]
    gm = {
        'e': [1.0, 2.0, 3.0],
        'f': [x for x in range(10)]
    }
    assert list(heterogeneous_stream(gm)) == [
        1.0, 0, 2.0, 1, 3.0, 2, 3, 4, 5, 6, 7, 8, 9
    ]



import venusian

from snosearch.configs import SearchConfigRegistry
from snovault.elasticsearch.searches.interfaces import SEARCH_CONFIG


def includeme(config):
    registry = config.registry
    registry[SEARCH_CONFIG] = SearchConfigRegistry()
    config.add_directive('register_search_config', register_search_config)


def register_search_config(config, name, factory):
    config.action(
        ('set-search-config', name),
        config.registry[SEARCH_CONFIG].register_from_func,
        args=(
            name,
            factory
        )
    )


def search_config(name, **kwargs):
    '''
    Register a custom search config by name.
    '''
    def decorate(config):
        def callback(scanner, factory_name, factory):
            scanner.config.register_search_config(name, factory)
        venusian.attach(config, callback, category='pyramid')
        return config
    return decorate



import pytest

# Use workbook fixture from BDD tests (including elasticsearch)
from .features.conftest import app_settings, app, workbook
from webob.multidict import MultiDict


def test_searchv2_view(workbook, testapp):
    r = testapp.get(
        '/search/?type=Snowflake&award=/awards/U41HG006992/&accession=SNOFL000LSQ&status=deleted'
    )
    assert r.json['title'] == 'Search'
    assert len(r.json['@graph']) == 1
    assert r.json['@graph'][0]['accession'] == 'SNOFL000LSQ'
    assert r.json['@graph'][0]['status'] == 'deleted'
    assert 'Snowflake' in r.json['@graph'][0]['@type']
    assert len(r.json['facets']) == 5
    assert r.json['@id'] == '/search/?type=Snowflake&award=/awards/U41HG006992/&accession=SNOFL000LSQ&status=deleted'
    assert r.json['@context'] == '/terms/'
    assert r.json['@type'] == ['Search']
    assert r.json['total'] == 1
    assert r.json['notification'] == 'Success'
    assert len(r.json['filters']) == 4
    assert r.status_code == 200
    assert r.json['clear_filters'] == '/search/?type=Snowflake'
    assert 'debug' not in r.json
    assert 'columns' in r.json
    assert 'sort' in r.json
    assert 'facets' in r.json


def test_searchv2_view_with_limit(workbook, testapp):
    r = testapp.get(
        '/search/?type=Snowflake&limit=5'
    )
    assert len(r.json['@graph']) == 5
    assert 'all' in r.json
    r = testapp.get(
        '/search/?type=Snowflake&limit=26'
    )
    assert len(r.json['@graph']) == 26
    assert 'all' in r.json
    r = testapp.get(
        '/search/?type=Snowflake&limit=all'
    )
    assert len(r.json['@graph']) == 35
    assert 'all' not in r.json
    r = testapp.get(
        '/search/?type=Snowflake&limit=35'
    )
    assert len(r.json['@graph']) == 35
    assert 'all' not in r.json
    r = testapp.get(
        '/search/?type=Snowflake&limit=100000'
    )
    assert len(r.json['@graph']) == 35
    assert 'all' not in r.json


def test_searchv2_view_with_limit_and_scan(workbook, testapp, mocker):
    from snosearch.queries import BasicSearchQueryFactoryWithFacets
    mocker.patch.object(BasicSearchQueryFactoryWithFacets, '_should_scan_over_results')
    BasicSearchQueryFactoryWithFacets._should_scan_over_results.return_value = True
    r = testapp.get(
        '/search/?type=Snowflake&limit=all'
    )
    assert len(r.json['@graph']) == 35
    r = testapp.get(
        '/search/?type=Snowflake&limit=30'
    )
    assert len(r.json['@graph']) == 30
    BasicSearchQueryFactoryWithFacets._should_scan_over_results.return_value = False
    r = testapp.get(
        '/search/?type=Snowflake&limit=all'
    )
    assert len(r.json['@graph']) == 25
    r = testapp.get(
        '/search/?type=Snowflake&limit=30'
    )
    assert len(r.json['@graph']) == 30


def test_searchv2_view_with_limit_zero(workbook, testapp):
    r = testapp.get(
        '/search/?type=Snowflake&limit=0'
    )
    assert len(r.json['@graph']) == 0
    assert 'all' in r.json
    assert r.json['total'] == 35


def test_searchv2_view_values(workbook, testapp):
    r = testapp.get(
        '/search/?status=released'
    )
    assert r.json['all'] == '/search/?status=released&limit=all'
    assert r.json['notification'] == 'Success'
    assert r.json['filters'][0] == {'field': 'status', 'remove': '/search/', 'term': 'released'}
    assert r.json['clear_filters'] == '/search/'


def test_searchv2_view_values_no_results(workbook, testapp):
    r = testapp.get(
        '/search/?status=current&type=Snowflake',
        status=404
    )
    assert r.json['notification'] == 'No results found'


def test_searchv2_view_values_malformed_query_string(workbook, testapp):
    r = testapp.get(
        '/search/?status=current&type=Snowflake&status=&format=json',
        status=404
    )
    assert r.json['notification'] == 'No results found'


def test_searchv2_view_values_regex_slash_escape(workbook, testapp):
    r = testapp.get(
        '/search/?searchTerm=/*'
    )
    assert r.json['total'] >= 32


def test_searchv2_view_values_bad_type(workbook, testapp):
    r = testapp.get(
        '/search/?status=released&type=Sno',
        status=400
    )
    assert r.json['description'] == "Invalid types: ['Sno']"
    r = testapp.get(
        '/search/?status=released&type=Sno&type=Flake',
        status=400
    )
    assert r.json['description'] == "Invalid types: ['Sno', 'Flake']"


def test_searchv2_view_values_item_wildcard(workbook, testapp):
    r = testapp.get(
        '/search/?type=*',
    )
    assert r.json['notification'] == 'Success'
    assert r.json['total'] >= 172


def test_searchv2_view_values_invalid_search_term(workbook, testapp):
    r = testapp.get(
        '/search/?searchTerm=[',
        status=404
    )
    r = testapp.get(
        '/search/?searchTerm=cherry^',
        status=200
    )
    assert r.json['total'] == 1


def test_searchv2_view_values_invalid_advanced_query(workbook, testapp):
    r = testapp.get(
        '/search/?advancedQuery=[',
        status=400
    )
    assert r.json['description'] == 'Invalid query: ([)'


def test_searchv2_view_values_reserved_characters_advanced_query(workbook, testapp):
    r = testapp.get(
        '/search/?searchTerm=cherry^',
        status=200
    )
    assert r.json['total'] == 1
    r = testapp.get(
        '/search/?searchTerm=cherry~',
        status=200
    )
    assert r.json['total'] == 1
    r = testapp.get(
        '/search/?searchTerm=/cherry',
        status=200
    )
    assert r.json['total'] == 1
    r = testapp.get(
        '/search/?searchTerm=/cherryp',
        status=404
    )
    assert r.json['total'] == 0
    r = testapp.get(
        '/search/?searchTerm=/cherry:',
        status=200
    )
    assert r.json['total'] == 1


def test_searchv2_view_embedded_frame(workbook, testapp):
    r = testapp.get(
        '/search/?type=Snowflake&frame=embedded'
    )
    assert r.json['@graph'][0]['lab']['name']


def test_searchv2_view_object_frame(workbook, testapp):
    r = testapp.get(
        '/search/?type=Snowflake&frame=object'
    )
    res = r.json['@graph'][0]
    assert all(
        [
            x in res
            for x in ['accession', '@type', '@id', 'status']
        ]
    )

def test_searchv2_view_debug_query(workbook, testapp):
    r = testapp.get(
        '/search/?type=Snowflake&debug=true'
    )
    assert 'debug' in r.json
    assert 'post_filter' in r.json['debug']['raw_query']


def test_searchv2_view_no_type(workbook, testapp):
    r = testapp.get('/search/')
    assert 'total' in r.json
    assert 'filters' in r.json
    assert len(r.json['filters']) == 0


def test_searchv2_view_no_type_debug(workbook, testapp):
    r = testapp.get('/search/?debug=true')
    assert not r.json['debug']['raw_query']['post_filter']['bool']


def test_searchv2_view_with_ranges(workbook, testapp):
    r = testapp.get(
        '/search/?type=Snowflake'
        '&date_created=gte:2013-03-15'
        '&date_created=lt:2013-04-01'
        '&date_created!=gt:2013-05-30'
        '&date_created!=lt:2013-03-01'
    )
    assert len(r.json['@graph']) == 2
    for snowflake in r.json['@graph']:
        assert snowflake['date_created'] == '2013-03-23'


def test_searchv2_raw_view_raw_response(workbook, testapp):
    r = testapp.get('/searchv2_raw/?type=Snowflake')
    assert 'hits' in r.json
    assert 'aggregations' in r.json
    assert '_shards' in r.json
    assert 'timed_out' in r.json
    assert 'took' in r.json


def test_searchv2_raw_view_raw_response_limit_all(workbook, testapp):
    r = testapp.get('/searchv2_raw/?type=*&limit=all')
    assert 'hits' in r.json
    assert len(r.json['hits']['hits']) > 170


def test_searchv2_quick_view_limit_all(workbook, testapp):
    r = testapp.get('/searchv2_quick/?type=*&limit=all')
    assert '@graph' in r.json
    assert len(r.json['@graph']) > 170
    assert '@id' in r.json['@graph'][0]
    assert len(r.json.keys()) == 1


def test_searchv2_quick_view_one_item(workbook, testapp):
    r = testapp.get(
        '/searchv2_quick/?type=Snowflake&award=/awards/U41HG006992/&accession=SNOFL000LSQ&status=deleted'
    )
    assert len(r.json['@graph']) == 1


def test_searchv2_quick_view_specify_field(workbook, testapp):
    r = testapp.get('/searchv2_quick/?type=Snowflake&field=@id')
    assert '@id' in r.json['@graph'][0]
    assert '@type' in r.json['@graph'][0]
    assert len(r.json['@graph'][0].keys()) == 2


def test_search_cached_facets_view(workbook, testapp):
    r = testapp.get(
        '/search-cached-facets/?type=Snowflake&award=/awards/U41HG006992/&accession=SNOFL000LSQ&status=deleted'
    )
    assert r.json['title'] == 'Search'
    assert len(r.json['@graph']) == 1
    assert r.json['@graph'][0]['accession'] == 'SNOFL000LSQ'
    assert r.json['@graph'][0]['status'] == 'deleted'
    assert 'Snowflake' in r.json['@graph'][0]['@type']
    assert len(r.json['facets']) == 5
    assert r.json['@id'] == '/search-cached-facets/?type=Snowflake&award=/awards/U41HG006992/&accession=SNOFL000LSQ&status=deleted'
    assert r.json['@context'] == '/terms/'
    assert r.json['@type'] == ['Search']
    assert r.json['total'] == 1
    assert r.json['notification'] == 'Success'
    assert len(r.json['filters']) == 4
    assert r.status_code == 200
    assert r.json['clear_filters'] == '/search-cached-facets/?type=Snowflake'
    assert 'debug' not in r.json
    assert 'columns' in r.json
    assert 'sort' in r.json
    assert 'facets' in r.json


def test_search_cached_facets_view_with_limit(workbook, testapp):
    r = testapp.get(
        '/search-cached-facets/?type=Snowflake&limit=5'
    )
    assert len(r.json['@graph']) == 5
    assert 'all' in r.json
    r = testapp.get(
        '/search-cached-facets/?type=Snowflake&limit=26'
    )
    assert len(r.json['@graph']) == 26
    assert 'all' in r.json
    r = testapp.get(
        '/search-cached-facets/?type=Snowflake&limit=all'
    )
    assert len(r.json['@graph']) == 35
    assert 'all' not in r.json
    r = testapp.get(
        '/search-cached-facets/?type=Snowflake&limit=35'
    )
    assert len(r.json['@graph']) == 35
    assert 'all' not in r.json
    r = testapp.get(
        '/search-cached-facets/?type=Snowflake&limit=100000'
    )
    assert len(r.json['@graph']) == 35
    assert 'all' not in r.json


def test_search_cached_facets_view_with_limit_and_scan(workbook, testapp, mocker):
    from snosearch.queries import BasicSearchQueryFactoryWithoutFacets
    mocker.patch.object(BasicSearchQueryFactoryWithoutFacets, '_should_scan_over_results')
    BasicSearchQueryFactoryWithoutFacets._should_scan_over_results.return_value = True
    r = testapp.get(
        '/search-cached-facets/?type=Snowflake&limit=all'
    )
    assert len(r.json['@graph']) == 35
    r = testapp.get(
        '/search-cached-facets/?type=Snowflake&limit=30'
    )
    assert len(r.json['@graph']) == 30
    BasicSearchQueryFactoryWithoutFacets._should_scan_over_results.return_value = False
    r = testapp.get(
        '/search-cached-facets/?type=Snowflake&limit=all'
    )
    assert len(r.json['@graph']) == 25
    r = testapp.get(
        '/search-cached-facets/?type=Snowflake&limit=30'
    )
    assert len(r.json['@graph']) == 30


def test_search_cached_facets_view_with_limit_zero(workbook, testapp):
    r = testapp.get(
        '/search-cached-facets/?type=Snowflake&limit=0'
    )
    assert len(r.json['@graph']) == 0
    assert 'all' in r.json
    assert r.json['total'] == 35


def test_search_cached_facets_view_values(workbook, testapp):
    r = testapp.get(
        '/search-cached-facets/?status=released'
    )
    assert r.json['all'] == '/search-cached-facets/?status=released&limit=all'
    assert r.json['notification'] == 'Success'
    assert r.json['filters'][0] == {'field': 'status', 'remove': '/search-cached-facets/', 'term': 'released'}
    assert r.json['clear_filters'] == '/search-cached-facets/'


def test_search_cached_facets_view_values_no_results(workbook, testapp):
    r = testapp.get(
        '/search-cached-facets/?status=current&type=Snowflake',
        status=404
    )
    assert r.json['notification'] == 'No results found'


def test_search_cached_facets_view_values_malformed_query_string(workbook, testapp):
    r = testapp.get(
        '/search-cached-facets/?status=current&type=Snowflake&status=&format=json',
        status=404
    )
    assert r.json['notification'] == 'No results found'


def test_search_cached_facets_view_values_invalid_search_term(workbook, testapp):
    r = testapp.get(
        '/search-cached-facets/?searchTerm=[',
        status=404
    )
    r = testapp.get(
        '/search-cached-facets/?searchTerm=cherry^',
        status=200
    )
    assert r.json['total'] == 1


def test_search_generator(workbook, threadlocals, dummy_request):
    from snosearch.parsers import ParamsParser
    from snovault.elasticsearch import ELASTIC_SEARCH
    from elasticsearch import Elasticsearch
    from types import GeneratorType
    dummy_request.environ['QUERY_STRING'] = (
        'type=*&limit=all'
    )
    dummy_request.registry[ELASTIC_SEARCH] = Elasticsearch(port=9201)
    from snowflakes.search_views import search_generator
    r = search_generator(dummy_request)
    assert '@graph' in r
    assert len(r.keys()) == 1
    assert isinstance(r['@graph'], GeneratorType)
    hits = [dict(h) for h in r['@graph']]
    assert len(hits) > 150
    assert '@id' in hits[0]


def test_search_generator_field_specified(workbook, threadlocals, dummy_request):
    from snosearch.parsers import ParamsParser
    from snovault.elasticsearch import ELASTIC_SEARCH
    from elasticsearch import Elasticsearch
    from types import GeneratorType
    dummy_request.environ['QUERY_STRING'] = (
        'type=Snowflake&field=@id&limit=5'
    )
    dummy_request.registry[ELASTIC_SEARCH] = Elasticsearch(port=9201)
    from snowflakes.search_views import search_generator
    r = search_generator(dummy_request)
    assert '@graph' in r
    assert len(r.keys()) == 1
    assert isinstance(r['@graph'], GeneratorType)
    hits = [dict(h) for h in r['@graph']]
    assert len(hits) == 5
    assert '@id' in hits[0]
    assert len(hits[0].keys()) == 2


def test_reportv2_view(workbook, testapp):
    r = testapp.get(
        '/report/?type=Snowflake&award=/awards/U41HG006992/&accession=SNOFL000LSQ&status=deleted'
    )
    assert r.json['title'] == 'Report'
    assert len(r.json['@graph']) == 1
    assert r.json['@graph'][0]['accession'] == 'SNOFL000LSQ'
    assert r.json['@graph'][0]['status'] == 'deleted'
    assert 'Snowflake' in r.json['@graph'][0]['@type']
    assert len(r.json['facets']) == 5
    assert r.json['@id'] == '/report/?type=Snowflake&award=/awards/U41HG006992/&accession=SNOFL000LSQ&status=deleted'
    assert r.json['@context'] == '/terms/'
    assert r.json['@type'] == ['Report']
    assert r.json['total'] == 1
    assert r.json['notification'] == 'Success'
    assert len(r.json['filters']) == 4
    assert r.status_code == 200
    assert r.json['clear_filters'] == '/report/?type=Snowflake'
    assert 'debug' not in r.json
    assert 'columns' in r.json
    assert 'non_sortable' in r.json
    assert 'sort' in r.json


def test_reportv2_response_with_search_term_type_only_clear_filters(workbook, testapp):
    r = testapp.get('/report/?type=Snowball&searchTerm=crisp')
    assert 'clear_filters' in r.json
    assert r.json['clear_filters'] == '/report/?type=Snowball'


def test_reportv2_view_with_limit(workbook, testapp):
    r = testapp.get(
        '/report/?type=Snowflake&limit=5'
    )
    assert len(r.json['@graph']) == 5
    assert 'all' in r.json
    r = testapp.get(
        '/report/?type=Snowflake&limit=26'
    )
    assert len(r.json['@graph']) == 26
    assert 'all' in r.json
    r = testapp.get(
        '/report/?type=Snowflake&limit=all'
    )
    assert len(r.json['@graph']) == 35
    assert 'all' not in r.json
    r = testapp.get(
        '/report/?type=Snowflake&limit=35'
    )
    assert len(r.json['@graph']) == 35
    assert 'all' not in r.json
    r = testapp.get(
        '/report/?type=Snowflake&limit=100000'
    )
    assert len(r.json['@graph']) == 35
    assert 'all' not in r.json


def test_reportv2_view_with_limit_zero(workbook, testapp):
    r = testapp.get(
        '/report/?type=Snowflake&limit=0'
    )
    assert len(r.json['@graph']) == 0
    assert 'all' in r.json
    assert r.json['total'] == 35


def test_reportv2_view_with_limit_zero_from_zero(workbook, testapp):
    r = testapp.get(
        '/report/?type=Snowflake&limit=0&from=0'
    )
    assert len(r.json['@graph']) == 0
    assert 'all' in r.json
    assert r.json['total'] == 35


def test_reportv2_view_values_bad_type(workbook, testapp):
    r = testapp.get(
        '/report/?status=released&type=Sno',
        status=400
    )
    assert r.json['description'] == "Invalid types: ['Sno']"
    r = testapp.get(
        '/report/?status=released&type=Sno&type=Flake',
        status=400
    )
    assert r.json['description'] == "Report view requires specifying a single type: [('type', 'Sno'), ('type', 'Flake')]"


def test_reportv2_view_values_single_subtype(workbook, testapp):
    r = testapp.get(
        '/report/?status=released&type=Item',
        status=400
    )
    assert 'Report view requires a type with no child types:' in r.json['description']


def test_reportv2_view_values_no_type(workbook, testapp):
    r = testapp.get(
        '/report/?status=released',
        status=400
    )
    assert r.json['description'] == 'Report view requires specifying a single type: []'


def test_report_cached_facets_view(workbook, testapp):
    r = testapp.get(
        '/report-cached-facets/?type=Snowflake&award=/awards/U41HG006992/&accession=SNOFL000LSQ&status=deleted'
    )
    assert r.json['title'] == 'Report'
    assert len(r.json['@graph']) == 1
    assert r.json['@graph'][0]['accession'] == 'SNOFL000LSQ'
    assert r.json['@graph'][0]['status'] == 'deleted'
    assert 'Snowflake' in r.json['@graph'][0]['@type']
    assert len(r.json['facets']) == 5
    assert r.json['@id'] == '/report-cached-facets/?type=Snowflake&award=/awards/U41HG006992/&accession=SNOFL000LSQ&status=deleted'
    assert r.json['@context'] == '/terms/'
    assert r.json['@type'] == ['Report']
    assert r.json['total'] == 1
    assert r.json['notification'] == 'Success'
    assert len(r.json['filters']) == 4
    assert r.status_code == 200
    assert r.json['clear_filters'] == '/report-cached-facets/?type=Snowflake'
    assert 'debug' not in r.json
    assert 'columns' in r.json
    assert 'non_sortable' in r.json
    assert 'sort' in r.json
    assert 'facets' in r.json


def test_matrixv2_raw_view_raw_response(workbook, testapp):
    r = testapp.get('/matrixv2_raw/?type=Snowball')
    assert 'hits' in r.json
    assert r.json['hits']['total'] >= 22
    assert len(r.json['hits']['hits']) == 0
    assert 'aggregations' in r.json
    assert 'x' in r.json['aggregations']
    assert 'y' in r.json['aggregations']
    assert 'snowflakes.type' in r.json['aggregations']['x']
    assert 'award.rfa' in r.json['aggregations']['y']
    assert '_shards' in r.json
    assert 'timed_out' in r.json
    assert 'took' in r.json


def test_matrixv2_raw_view_no_matrix_defined(workbook, testapp):
    r = testapp.get(
        '/matrixv2_raw/?type=Snowflake',
        status=400
    )
    assert r.json['description'] == 'Item type does not have requested view defined: {}'


def test_matrixv2_raw_view_invalid_type(workbook, testapp):
    r = testapp.get(
        '/matrixv2_raw/?type=Sno',
        status=400
    )
    assert r.json['description'] == "Invalid types: ['Sno']"


def test_matrixv2_raw_view_mutiple_types(workbook, testapp):
    r = testapp.get(
        '/matrixv2_raw/?type=Snowflake&type=Snowball',
        status=400
    )
    assert 'Matrix view requires specifying a single type' in r.json['description']


def test_matrixv2_raw_view_no_types(workbook, testapp):
    r = testapp.get(
        '/matrixv2_raw/',
        status=400
    )
    assert r.json['description'] == 'Matrix view requires specifying a single type: []'


def test_matrixv2_response(workbook, testapp):
    r = testapp.get('/matrix/?type=Snowball')
    assert 'aggregations' not in r.json
    assert 'facets' in r.json
    assert 'total' in r.json
    assert r.json['title'] == 'Matrix'
    assert r.json['@type'] == ['Matrix']
    assert r.json['clear_filters'] == '/matrix/?type=Snowball'
    assert r.json['filters'] == [{'term': 'Snowball', 'remove': '/matrix/', 'field': 'type'}]
    assert r.json['@id'] == '/matrix/?type=Snowball'
    assert r.json['total'] >= 22
    assert r.json['notification'] == 'Success'
    assert r.json['title'] == 'Matrix'
    assert 'facets' in r.json
    assert r.json['@context'] == '/terms/'
    assert 'matrix' in r.json
    assert r.json['matrix']['x']['group_by'] == 'snowflakes.type'
    assert r.json['matrix']['y']['group_by'] == ['award.rfa', 'lab.title']
    assert 'buckets' in r.json['matrix']['y']['award.rfa']
    assert 'key' in r.json['matrix']['y']['award.rfa']['buckets'][0]
    assert 'lab.title' in r.json['matrix']['y']['award.rfa']['buckets'][0]
    assert 'search_base' in r.json
    assert r.json['search_base'] == '/search/?type=Snowball'


def test_matrixv2_response_with_search_term_type_only_clear_filters(workbook, testapp):
    r = testapp.get('/matrix/?type=Snowball&searchTerm=crisp')
    assert 'clear_filters' in r.json
    assert r.json['clear_filters'] == '/matrix/?type=Snowball'


def test_matrixv2_response_debug(workbook, testapp):
    r = testapp.get('/matrix/?type=Snowball&debug=true')
    assert 'debug' in r.json


def test_matrixv2_response_no_results(workbook, testapp):
    r = testapp.get(
        '/matrix/?type=Snowball&status=no_status',
        status=404
    )
    assert r.json['notification'] == 'No results found'


def test_missing_matrix_response(workbook, testapp):
    r = testapp.get('/missing_matrix/?type=Snowball')
    assert 'aggregations' not in r.json
    assert 'facets' in r.json
    assert 'total' in r.json
    assert r.json['title'] == 'Matrix'
    assert r.json['@type'] == ['Matrix']
    assert r.json['clear_filters'] == '/missing_matrix/?type=Snowball'
    assert r.json['filters'] == [{'term': 'Snowball', 'remove': '/missing_matrix/', 'field': 'type'}]
    assert r.json['@id'] == '/missing_matrix/?type=Snowball'
    assert r.json['total'] >= 22
    assert r.json['notification'] == 'Success'
    assert r.json['title'] == 'Matrix'
    assert 'facets' in r.json
    assert r.json['@context'] == '/terms/'
    assert 'matrix' in r.json
    assert r.json['matrix']['x']['group_by'] == 'snowflakes.type'
    assert r.json['matrix']['y']['group_by'] == ['award.rfa', ['lab.not_a_real_value', 'some_lab']]
    assert 'buckets' in r.json['matrix']['y']['award.rfa']
    assert 'key' in r.json['matrix']['y']['award.rfa']['buckets'][0]
    assert 'lab.not_a_real_value' in r.json['matrix']['y']['award.rfa']['buckets'][0]
    assert r.json['matrix']['y']['award.rfa']['buckets'][0]['lab.not_a_real_value']['buckets'][0]['key'] == 'some_lab'
    assert 'search_base' in r.json
    assert r.json['search_base'] == '/search/?type=Snowball'


def test_summaryv2_response(workbook, testapp):
    r = testapp.get('/summary/?type=Snowball')
    assert 'aggregations' not in r.json
    assert 'facets' in r.json
    assert 'total' in r.json
    assert r.json['title'] == 'Summary'
    assert r.json['@type'] == ['Summary']
    assert r.json['clear_filters'] == '/summary/?type=Snowball'
    assert r.json['filters'] == [{'term': 'Snowball', 'remove': '/summary/', 'field': 'type'}]
    assert r.json['@id'] == '/summary/?type=Snowball'
    assert r.json['total'] >= 22
    assert r.json['notification'] == 'Success'
    assert r.json['title'] == 'Summary'
    assert 'facets' in r.json
    assert r.json['@context'] == '/terms/'
    assert 'matrix' in r.json
    assert r.json['matrix']['x']['group_by'] == 'status'
    assert r.json['matrix']['y']['group_by'] == ['snowflakes.type']
    assert 'buckets' in r.json['matrix']['y']['snowflakes.type']
    assert 'key' in r.json['matrix']['y']['snowflakes.type']['buckets'][0]
    assert 'status' in r.json['matrix']['y']['snowflakes.type']['buckets'][0]
    assert 'search_base' in r.json
    assert r.json['search_base'] == '/search/?type=Snowball'


def test_collection_listing_es_view(workbook, testapp):
    r = testapp.get(
        '/snowflakes/'
    )
    assert '@graph' in r.json
    assert '@id' in r.json
    assert 'facets' in r.json
    assert 'filters' in r.json
    assert 'all' in r.json
    assert 'columns' in r.json
    assert 'clear_filters' in r.json
    assert r.json['clear_filters'] == '/search/?type=Snowflake'
    assert r.json['@type'] == ['SnowflakeCollection', 'Collection']
    assert r.json['@context'] == '/terms/'


def test_collection_listing_es_view_item(workbook, testapp):
    r = testapp.get(
        '/Snowflake'
    )
    r = r.follow()
    assert '@graph' in r.json
    assert '@id' in r.json
    assert 'facets' in r.json
    assert 'filters' in r.json
    assert r.json['@type'] == ['SnowflakeCollection', 'Collection']
    assert r.json['@context'] == '/terms/'


def test_collection_listing_db_view(workbook, testapp):
    r = testapp.get(
        '/snowflakes/?datastore=database'
    )
    assert '@graph' in r.json
    assert '@id' in r.json
    assert 'facets' not in r.json
    assert 'filters' not in r.json
    assert r.json['@type'] == ['SnowflakeCollection', 'Collection']
    assert r.json['@context'] == '/terms/'
    assert r.json['description'] == 'Listing of Snowflakes'


def test_auditv2_response(workbook, testapp):
    r = testapp.get('/audit/?type=Snowball')
    assert 'aggregations' not in r.json
    assert 'facets' in r.json
    assert 'total' in r.json
    assert r.json['title'] == 'Audit'
    assert r.json['@type'] == ['Audit']
    assert r.json['clear_filters'] == '/audit/?type=Snowball'
    assert r.json['filters'] == [{'term': 'Snowball', 'remove': '/audit/', 'field': 'type'}]
    assert r.json['@id'] == '/audit/?type=Snowball'
    assert r.json['total'] >= 22
    assert r.json['notification'] == 'Success'
    assert 'facets' in r.json
    assert r.json['@context'] == '/terms/'
    assert 'matrix' in r.json
    assert r.json['matrix']['x']['group_by'] == 'snowflakes.type'
    assert r.json['matrix']['audit.ERROR.category']
    assert r.json['matrix']['audit.WARNING.category']
    assert r.json['matrix']['audit.NOT_COMPLIANT.category']
    assert r.json['matrix']['audit.INTERNAL_ACTION.category']
    assert 'search_base' in r.json
    assert r.json['search_base'] == '/search/?type=Snowball'


def test_auditv2_response_with_search_term_type_only_clear_filters(workbook, testapp):
    r = testapp.get('/audit/?type=Snowball&searchTerm=crisp')
    assert 'clear_filters' in r.json
    assert r.json['clear_filters'] == '/audit/?type=Snowball'


def test_auditv2_response_debug(workbook, testapp):
    r = testapp.get('/audit/?type=Snowball&debug=true')
    assert 'debug' in r.json


def test_auditv2_response_no_results(workbook, testapp):
    r = testapp.get(
        '/audit/?type=Snowball&status=no_status',
        status=404
    )
    assert r.json['notification'] == 'No results found'


def test_auditv2_view_no_matrix_defined(workbook, testapp):
    r = testapp.get(
        '/audit/?type=Snowflake',
        status=400
    )
    assert r.json['description'] == 'Item type does not have requested view defined: {}'


def test_top_hits_raw_view_response(workbook, testapp):
    r = testapp.get('/top_hits_raw/?searchTerm=cherry')
    assert not r.json['hits']['hits']
    assert 'aggregations' in r.json
    assert '_shards' in r.json
    assert 'timed_out' in r.json
    assert 'took' in r.json
    assert len(r.json['aggregations']['types']['types']['buckets']) == 1
    assert r.json['aggregations']['types']['types']['buckets'][0]['key'] == 'Lab'



from pyramid.paster import get_app
import logging
from webtest import TestApp

EPILOG = __doc__


def run(app, first_name, last_name, email, lab):
    environ = {
        'HTTP_ACCEPT': 'application/json',
        'REMOTE_USER': 'TEST',
    }
    testapp = TestApp(app, environ)
    testapp.post_json('/user', {
        'first_name': first_name,
        'last_name': last_name,
        'email': email,
        'lab': lab,
        'groups': ['admin']
    })


def main():
    ''' Creates an admin user '''

    import argparse
    parser = argparse.ArgumentParser(
        description="Creates an admin user", epilog=EPILOG,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument('--first-name', default='Admin', help="First name")
    parser.add_argument('--last-name', default='Test', help="Last name")
    parser.add_argument('--email', default='admin_test@example.org', help="E-mail")
    parser.add_argument('--lab', default='/labs/j-michael-cherry/', help="Lab")
    parser.add_argument('--app-name', help="Pyramid app name in configfile")
    parser.add_argument('config_uri', help="path to configfile")
    args = parser.parse_args()

    logging.basicConfig()
    options = {
        'embed_cache.capacity': '5000',
        'indexer': 'true',
    }
    app = get_app(args.config_uri, args.app_name, options)

    logging.getLogger('encoded').setLevel(logging.DEBUG)
    return run(app, args.first_name, args.last_name, args.email, args.lab)


if __name__ == '__main__':
    main()



''''
Test the functionality of SimpleUuidServer and SimpleUuidWorker
in the context of Indexer and MPIndexer
'''
from unittest import (
    TestCase,
    mock,
)

from snovault.elasticsearch.simple_queue import (
    SimpleUuidServer,
    SimpleUuidWorker,
)


MOCK_TIME = 123456.654321
QUEUE_RESET_MSG_BASE = 'Queue Server cannot reset. '


class TestSimpleServer(TestCase):
    '''Test Simple Uuid Server'''
    # pylint: disable=too-many-public-methods
    members = [
        ('_queue_options', dict),
        ('_start_us', int),
        ('queue_id', str),
        ('_uuids', list),
        ('_uuid_count', int),
        ('_worker_conns', dict),
        ('_worker_results', dict),
        ('_errors', list),
    ]
    meth_func = [
        # Errors
        'add_errors',
        '_has_errors',
        'pop_errors',
        # Worker
        '_get_blank_worker',
        '_add_worker_conn',
        'get_worker_conns',
        'get_worker_conn_count',
        '_get_worker_id',
        '_get_worker_queue',
        'get_worker',
        'update_worker_conn',
        'save_work_results',
        # Uuids
        'has_uuids',
        '_get_uuid',
        'get_uuids',
        '_load_uuid',
        'load_uuids',
        # Run
        'is_indexing',
        'update_finished',
        'close_indexing',
    ]

    @classmethod
    @mock.patch('time.time', mock.MagicMock(return_value=MOCK_TIME))
    def setUpClass(cls):
        '''Create a simple worker to be used most tests'''
        cls.load_uuids = list(range(10, 99))
        cls.queue_options = {
            'processes': 1,
            'chunk_size': 1,
            'batch_size': 10,
        }
        cls.mock_time_us = int(MOCK_TIME * 1000000)
        cls.queue_server = SimpleUuidServer(cls.queue_options)
        cls.queue_worker = None

    def setUp(self):
        '''Setup before each method call'''
        self.queue_worker = self.queue_server.get_worker()

    def tearDown(self):
        """Reset for each test"""
        self.queue_worker = None
        # pylint: disable=protected-access
        self.queue_server._uuids = []
        self.queue_server._uuid_count = 0
        self.queue_server._worker_conns = {}
        self.queue_server._worker_results = {}
        self.queue_server._errors = []

    # private/protected
    def test_init_dir(self):
        '''Test SimpleUuidServer has expected function and variables'''
        dir_items = [
            item
            for item in dir(self.queue_server)
            if item[0:2] != '__'
        ]
        member_names = [name for name, _ in self.members]
        for item in dir_items:
            if item in member_names:
                continue
            if item in self.meth_func:
                continue
            raise AssertionError('Member or method not being tested:%s' % item)
        for item, item_type in self.members:
            self.assertTrue(hasattr(self.queue_server, item))
            self.assertIsInstance(
                getattr(self.queue_server, item),
                item_type
            )
        for item in self.meth_func:
            self.assertTrue(hasattr(self.queue_server, item))
            self.assertTrue(
                str(type(getattr(self.queue_server, item))) in [
                    "<class 'method'>",
                    "<class 'function'>"
                ]
            )

    def test_init_basic(self):
        '''Test _get_worker_id'''
        self.assertEqual(str(self.mock_time_us), self.queue_server.queue_id)
        # pylint: disable=protected-access
        self.assertEqual(self.mock_time_us, self.queue_server._start_us)
        self.assertDictEqual(
            self.queue_options,
            self.queue_server._queue_options
        )

    # Errors
    def test_errors(self):
        '''Test add_errors'''
        worker_id = self.queue_worker.worker_id
        errors = ['e1', 'e2', 'e3']
        errors_to_add = [{'err': error} for error in errors]
        errors_added = self.queue_server.add_errors(worker_id, errors_to_add)
        self.assertEqual(errors_added, len(errors_to_add))
        error_msgs = []
        # pylint: disable=protected-access
        for error in self.queue_server._errors:
            self.assertTrue(error['worker_id'] == worker_id)
            error_msgs.append(error['err'])
        errors.sort()
        error_msgs.sort()
        self.assertListEqual(errors, error_msgs)

    def test_has_errors_true(self):
        '''Test _has_errors return true when not empty'''
        # pylint: disable=protected-access
        self.queue_server._errors = ['e']
        self.assertTrue(self.queue_server._has_errors())

    def test_has_errors_false(self):
        '''Test _has_errors returns false when empty'''
        # pylint: disable=protected-access
        self.assertFalse(self.queue_server._has_errors())

    def test_pop_errors(self):
        '''Test pop_errors'''
        # pylint: disable=protected-access
        self.assertFalse(self.queue_server._has_errors())

    # Worker
    def test_get_blank_worker(self):
        '''Test _get_blank_worker'''
        # pylint: disable=protected-access
        self.assertDictEqual(
            self.queue_server._get_blank_worker(),
            {'uuid_cnt': 0, 'get_cnt': 0}
        )

    def test_add_worker_conn(self):
        '''Test _add_worker_conn'''
        worker_id = 'some-id'
        # pylint: disable=protected-access
        self.queue_server._add_worker_conn(worker_id)
        self.assertTrue(worker_id in self.queue_server._worker_results)
        self.assertTrue(worker_id in self.queue_server._worker_conns)

    def test_get_worker_conns(self):
        '''Test get_worker_conns'''
        worker_id = self.queue_worker.worker_id
        worker_conns = self.queue_server.get_worker_conns()
        self.assertTrue(len(worker_conns) == 1)
        self.assertTrue(worker_id in worker_conns)
        # pylint: disable=protected-access
        self.assertDictEqual(
            worker_conns[worker_id],
            self.queue_server._get_blank_worker(),
        )

    def test_get_worker_conn_count(self):
        '''Test get_worker_conn_count'''
        self.assertTrue(
            self.queue_server.get_worker_conn_count(),
            1
        )

    def test_get_worker_id(self):
        '''Test _get_worker_id'''
        # Plus two since a worker is added in setUp
        id_time_us = self.mock_time_us + 2
        # pylint: disable=protected-access
        self.assertTrue(
            self.queue_server._get_worker_id() == str(id_time_us)
        )
        _ = self.queue_server._get_worker_id()
        # The function does not add the worker to the conn
        # so the id should not change
        self.assertTrue(
            self.queue_server._get_worker_id() == str(id_time_us)
        )

    def test_get_worker_queue(self):
        '''Test _get_worker_queue'''
        # pylint: disable=protected-access
        self.assertTrue(self.queue_server is self.queue_server._get_worker_queue())

    def test_get_worker_return(self):
        '''Test get_worker return value'''
        self.assertIsInstance(
            self.queue_server.get_worker(),
            SimpleUuidWorker
        )

    def test_get_worker_conn(self):
        '''Test get_worker worker connection entry'''
        worker = self.queue_server.get_worker()
        # pylint: disable=protected-access
        self.assertEqual(2, len(self.queue_server._worker_conns))
        worker_conn = self.queue_server._worker_conns.get(worker.worker_id)
        self.assertFalse(worker_conn is None)
        self.assertDictEqual(
            worker_conn,
            {
                'uuid_cnt': 0,
                'get_cnt': 0,
            }
        )

    def test_get_worker_many(self):
        '''Test get_worker with 2 workers'''
        worker_2 = self.queue_server.get_worker()
        worker_3 = self.queue_server.get_worker()
        # pylint: disable=protected-access
        self.assertEqual(len(self.queue_server._worker_conns), 3)
        self.assertTrue(worker_2.worker_id == str(self.mock_time_us + 2))
        self.assertTrue(worker_3.worker_id == str(self.mock_time_us + 3))

    def test_update_worker_conn(self):
        '''Test update_worker_conn'''
        update_uuid_cnt = 9
        update_get_cnt = 3
        worker_id = self.queue_worker.worker_id
        # pylint: disable=protected-access
        worker_conn = self.queue_server._worker_conns[worker_id]
        self.queue_server.update_worker_conn(
            worker_id,
            update_uuid_cnt,
            update_get_cnt,
        )
        self.assertDictEqual(
            worker_conn,
            {
                'uuid_cnt': update_uuid_cnt,
                'get_cnt': update_get_cnt
            }
        )

    def test_save_work_results(self):
        '''Test save_work_results'''
        worker_id = self.queue_worker.worker_id
        save_res = {'res': 'some-res'}
        self.queue_server.save_work_results(worker_id, save_res)
        # pylint: disable=protected-access
        self.assertTrue(len(self.queue_server._worker_results) == 1)
        self.assertTrue(worker_id in self.queue_server._worker_results)
        self.assertDictEqual(
            self.queue_server._worker_results[worker_id][0],
            save_res,
        )

    # Uuids
    def test_has_uuids(self):
        '''Test has_uuids'''
        # pylint: disable=protected-access
        self.queue_server._uuids = [1, 2, 3]
        self.assertTrue(self.queue_server.has_uuids())

    def test_has_uuids_empty(self):
        '''Test has_uuids when empty'''
        self.assertFalse(self.queue_server.has_uuids())

    def test_get_uuid(self):
        '''Test _get_uuid'''
        # pylint: disable=protected-access
        self.queue_server._uuids = list(reversed(self.load_uuids))
        uuid = self.queue_server._get_uuid()
        self.assertEqual(uuid, self.load_uuids[0])

    def test_get_uuid_empty(self):
        '''Test _get_uuid with empty'''
        # pylint: disable=protected-access
        self.assertIsNone(self.queue_server._get_uuid())

    def test_get_uuids(self):
        '''Test get_uuids'''
        uuid_cnt = 11
        # pylint: disable=protected-access
        self.queue_server._uuids = list(reversed(self.load_uuids))
        expected_uuids = self.load_uuids[0:uuid_cnt]
        uuids = self.queue_server.get_uuids(uuid_cnt)
        self.assertListEqual(expected_uuids, uuids)

    def test_get_uuids_all(self):
        '''Test get_uuids all with new worker connection'''
        uuid_cnt = -1
        # pylint: disable=protected-access
        self.queue_server._uuids = list(reversed(self.load_uuids))
        uuids = self.queue_server.get_uuids(uuid_cnt)
        self.load_uuids.sort()
        uuids.sort()
        self.assertListEqual(self.load_uuids, uuids)

    def test_load_uuid(self):
        '''Test _load_uuid'''
        # pylint: disable=protected-access
        self.queue_server._uuids = list(reversed(self.load_uuids))
        new_uuid = 'a-uuid'
        self.queue_server._load_uuid(new_uuid)
        self.assertEqual(new_uuid, self.queue_server._uuids[0])
        self.assertEqual(len(self.load_uuids) + 1, len(self.queue_server._uuids))

    def test_load_uuids(self):
        '''Test load_uuids'''
        len_load_uuids = self.queue_server.load_uuids(self.load_uuids.copy())
        self.assertEqual(len(self.load_uuids), len_load_uuids)
        # pylint: disable=protected-access
        self.assertListEqual(
            self.queue_server._uuids,
            list(reversed(self.load_uuids))
        )

    def test_load_uuids_full(self):
        '''Test load_uuids when has uuids already'''
        # pylint: disable=protected-access
        previous_uuids = ['1']
        self.queue_server._uuids = previous_uuids
        len_load_uuids = self.queue_server.load_uuids(self.load_uuids.copy())
        self.assertListEqual(
            previous_uuids,
            self.queue_server._uuids,
        )
        self.assertIsNone(len_load_uuids)

    # Run
    def test_is_indexing_false_all(self):
        '''
        Test is_indexing is false
        -No uuids, no errors, and worker uuid cnt is 0
        '''
        self.assertFalse(self.queue_server.is_indexing())

    def test_is_indexing_true_uuidcnt(self):
        '''
        Test is_indexing is true when work uuid cnt is not 0
        '''
        worker_id = self.queue_worker.worker_id
        # pylint: disable=protected-access
        self.queue_server._worker_conns[worker_id]['uuid_cnt'] = 1
        self.assertTrue(self.queue_server.is_indexing())

    def test_is_indexing_true_haserrors(self):
        '''
        Test is_indexing is true when has errors is true
        '''
        # pylint: disable=protected-access
        self.queue_server._errors = [1]
        self.assertTrue(self.queue_server.is_indexing())

    def test_is_indexing_true_hasuuids(self):
        '''
        Test is_indexing is true when has uuids is true
        '''
        # pylint: disable=protected-access
        self.queue_server._uuids = [1]
        self.assertTrue(self.queue_server.is_indexing())

    def test_update_finished_nowrkid(self):
        '''Test update_finished with bad worker id'''
        worker_id = 'bad-id'
        uuid_cnt = 11
        results = {'errors': [], 'successes': uuid_cnt}
        res_msg = self.queue_server.update_finished(worker_id, results)
        self.assertEqual(
            res_msg,
            "Worker id(%s) DNE." % worker_id
        )

    def test_update_finished_cntoffsucc(self):
        '''Test update_finished with success count off'''
        worker_id = self.queue_worker.worker_id
        uuid_cnt = 11
        # pylint: disable=protected-access
        self.queue_server._worker_conns[worker_id]['uuid_cnt'] = uuid_cnt
        results = {'errors': [], 'successes': uuid_cnt - 1}
        res_msg = self.queue_server.update_finished(worker_id, results)
        self.assertEqual(
            res_msg,
            'Queue Server cannot close worker (%s)' % worker_id
        )

    def test_update_finished_cntofferr(self):
        '''Test update_finished with errors count off'''
        worker_id = self.queue_worker.worker_id
        uuid_cnt = 11
        # pylint: disable=protected-access
        self.queue_server._worker_conns[worker_id]['uuid_cnt'] = uuid_cnt
        errors = [{'err': str(cnt)} for cnt in range(uuid_cnt + 1)]
        results = {'errors': errors, 'successes': 0}
        res_msg = self.queue_server.update_finished(worker_id, results)
        self.assertEqual(
            res_msg,
            'Queue Server cannot close worker (%s)' % worker_id
        )

    def test_update_finished_cntoffboth(self):
        '''Test update_finished with errors and success count off'''
        worker_id = self.queue_worker.worker_id
        uuid_cnt = 11
        # pylint: disable=protected-access
        self.queue_server._worker_conns[worker_id]['uuid_cnt'] = uuid_cnt
        errors = [{'err': str(cnt)} for cnt in range(uuid_cnt - 3)]
        results = {'errors': errors, 'successes': 0}
        res_msg = self.queue_server.update_finished(worker_id, results)
        self.assertEqual(
            res_msg,
            'Queue Server cannot close worker (%s)' % worker_id
        )

    def test_update_finished_okay_succ(self):
        '''Test update_finished with success okay msg'''
        worker_id = self.queue_worker.worker_id
        uuid_cnt = 11
        # pylint: disable=protected-access
        self.queue_server._worker_conns[worker_id]['uuid_cnt'] = uuid_cnt
        results = {'errors': [], 'successes': uuid_cnt}
        res_msg = self.queue_server.update_finished(worker_id, results)
        self.assertEqual(
            res_msg,
            'Okay'
        )

    def test_update_finished_okayerrs(self):
        '''Test update_finished with errors okay msg'''
        worker_id = self.queue_worker.worker_id
        uuid_cnt = 11
        # pylint: disable=protected-access
        self.queue_server._worker_conns[worker_id]['uuid_cnt'] = uuid_cnt
        errors = [{'err': str(cnt)} for cnt in range(uuid_cnt)]
        results = {'errors': errors, 'successes': 0}
        res_msg = self.queue_server.update_finished(worker_id, results)
        self.assertEqual(
            res_msg,
            'Okay'
        )

    def test_update_finished_okerrsucc(self):
        '''Test update_finished with errors and success okay msg'''
        worker_id = self.queue_worker.worker_id
        uuid_cnt = 11
        err_cnt = 4
        # pylint: disable=protected-access
        self.queue_server._worker_conns[worker_id]['uuid_cnt'] = uuid_cnt
        errors = [{'err': str(cnt)} for cnt in range(err_cnt)]
        results = {'errors': errors, 'successes': uuid_cnt - err_cnt}
        res_msg = self.queue_server.update_finished(worker_id, results)
        self.assertEqual(
            res_msg,
            'Okay'
        )


class TestSimpleUuidWorker(TestCase):
    '''Test Simple Uuid worker'''
    members = [
        ('queue_options', dict),
        ('worker_id', str),
        ('_queue', SimpleUuidServer),
        ('is_running', bool),
        ('get_cnt', int),
        ('uuid_cnt', int),
        ('processes', int),
        ('chunk_size', int),
    ]
    methods = [
        'get_uuids',
        'update_finished',
    ]

    @classmethod
    @mock.patch('time.time', mock.MagicMock(return_value=MOCK_TIME))
    def setUpClass(cls):
        '''Create a simple worker to be used most tests'''
        queue_options = {
            'processes': 1,
            'chunk_size': 1,
            'batch_size': 10,
        }
        cls.queue_options = queue_options
        cls.mock_server = SimpleUuidServer(queue_options)
        uuids_loaded = [str(uuid) for uuid in range(100, 150)]
        cls.mock_return_uuids = uuids_loaded[0:queue_options['batch_size']]
        cls.mock_server.get_uuids = mock.MagicMock(
            return_value=cls.mock_return_uuids
        )
        cls.mock_server.update_finished = mock.MagicMock(return_value=None)
        cls.queue_worker = cls.mock_server.get_worker()

    def test_init_dir(self):
        '''Test SimpleUuidServer has expected function and variables'''
        dir_items = [
            item
            for item in dir(self.queue_worker)
            if item[0:2] != '__'
        ]
        member_names = [name for name, _ in self.members]
        for item in dir_items:
            if item in member_names:
                continue
            if item in self.methods:
                continue
            raise AssertionError('Member or method not being tested:%s' % item)
        for item, item_type in self.members:
            self.assertTrue(hasattr(self.queue_worker, item))
            self.assertIsInstance(
                getattr(self.queue_worker, item),
                item_type
            )
        for item in self.methods:
            self.assertTrue(hasattr(self.queue_worker, item))
            self.assertEqual(
                str(type(getattr(self.queue_worker, item))),
                "<class 'method'>"
            )

    def test_init_basic(self):
        '''Test a basic initialization and _get_worker_id'''
        self.assertEqual(
            self.queue_options['processes'],
            self.queue_worker.queue_options['processes']
        )
        self.assertEqual(
            self.queue_options['chunk_size'],
            self.queue_worker.queue_options['chunk_size']
        )
        self.assertEqual(
            self.queue_options['batch_size'],
            self.queue_worker.queue_options['batch_size']
        )
        # pylint: disable=protected-access
        self.assertTrue(self.mock_server is self.queue_worker._queue)
        worker_id = str(int(MOCK_TIME * 1000000) + 1)
        self.assertEqual(worker_id, self.queue_worker.worker_id)

    # Uuids
    def test_get_uuids_called(self):
        '''Test worker get_uuids calls server get_uuids'''
        self.queue_worker.get_uuids()
        self.assertEqual(1, self.mock_server.get_uuids.call_count)
        self.mock_server.get_uuids.assert_called_with(
            self.queue_options['batch_size'],
            get_all=False,
        )
        self.assertListEqual(
            self.mock_return_uuids,
            self.mock_server.get_uuids.return_value
        )

    # Run
    def test_update_finished_called(self):
        '''Test worker update_finished calls server update_finished'''
        results = {}
        self.queue_worker.update_finished(results)
        self.assertEqual(1, self.mock_server.update_finished.call_count)
        self.mock_server.update_finished.assert_called_with(
            self.queue_worker.worker_id,
            results,
        )
        self.assertIsNone(self.mock_server.update_finished.return_value)



# https://code.google.com/p/modwsgi/wiki/RegisteringCleanupCode


class Generator2:
    def __init__(self, iterable, callback, environ):
        self.__iterable = iterable
        self.__callback = callback
        self.__environ = environ

    def __iter__(self):
        for item in self.__iterable:
            yield item

    def close(self):
        try:
            if hasattr(self.__iterable, 'close'):
                self.__iterable.close()
        finally:
            self.__callback(self.__environ)


class ExecuteOnCompletion2:
    def __init__(self, application, callback):
        self.__application = application
        self.__callback = callback

    def __call__(self, environ, start_response):
        try:
            result = self.__application(environ, start_response)
        except:
            self.__callback(environ)
            raise
        return Generator2(result, self.__callback, environ)


import logging
import psutil
import humanfriendly


def rss_checker(rss_limit=None):
    log = logging.getLogger(__name__)
    process = psutil.Process()

    def callback(environ):
        rss = process.memory_info().rss
        if rss_limit and rss > rss_limit:
            msg = "Restarting process. Memory usage exceeds limit of %d: %d"
            log.error(msg, rss_limit, rss)
            process.kill()

    return callback


def filter_app(app, global_conf, rss_limit=None):
    if rss_limit is not None:
        rss_limit = humanfriendly.parse_size(rss_limit)

    callback = rss_checker(rss_limit)
    return ExecuteOnCompletion2(app, callback)



# This has to be part of a plugin to support adding the command line option
import argparse


class AppendInt2(argparse._AppendAction):
    def __call__(self, parser, namespace, values, option_string=None):
        values = (values[0], int(values[1]))
        return super(AppendInt2, self).__call__(parser, namespace, values, option_string)


def pytest_addoption(parser):
    parser.addoption('--browser-arg', nargs=2, dest='browser_args', action='append', type='string')
    parser.addoption('--browser-arg-int', nargs=2, dest='browser_args', action=AppendInt2, type='string')
    parser.addoption('--chrome-options', action='store', type='string')
    parser.addoption('--wsgi-arg', nargs=2, dest='wsgi_args', action='append', type='string')



"""
Uuid Queue Module Adapter

- QueueAdapter Class allows access to all queue types
defined in QueueTypes through a set of standard methods.
- All queues in ./queues should adhere to QueueAdapter standards.
- Adapter queue has a server and a worker.
- Another important object is the meta data needed to run the queue.
"""
from .adapter_queue import QueueAdapter
from .adapter_queue import QueueTypes



import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.rst')).read()
CHANGES = open(os.path.join(here, 'CHANGES.rst')).read()
VERSION = "1.0.61"


INSTALL_REQUIRES = [
    "Pillow==7.0.0",
    "SQLAlchemy==1.3.13",
    "WebOb==1.8.6",
    "WebTest==2.0.35",
    "bcrypt==3.1.7",
    "boto3==1.11.9",
    "botocore==1.14.9",
    "elasticsearch-dsl==5.4.0",
    "elasticsearch==5.4.0",
    "future==0.18.2",
    "jsonschema_serialize_fork==2.1.1",
    "lucenequery==0.1",
    "passlib==1.7.2",
    "psutil==5.6.7",
    "psycopg2==2.8.4",
    "pyramid-localroles==0.1",
    "pyramid-multiauth==0.9.0",
    "pyramid-tm==2.4",
    "pyramid==1.10.4",
    "pyramid_retry==2.1.1",
    "python-magic==0.4.15",
    "pytz==2019.3",
    "rdflib-jsonld==0.4.0",
    "rdflib==4.2.2",
    "redis==3.5.3",
    "requests==2.22.0",
    "simplejson==3.17.0",
    "snovault-search==1.0.3",
    "transaction==3.0.0",
    "venusian==3.0.0",
    "zope.interface==5.3.0",
    "zope.sqlalchemy==1.2",
]

EXTRAS_REQUIRE = {
    "tests": [
        "pytest==5.3.2",
        "pytest-bdd==3.2.1",
        "pytest-mock==2.0.0",
        "pytest-splinter==2.0.1",
        "pytest-exact-fixtures==0.3",
        "pytest-instafail==0.4.1.post0",
        "pytest-timeout==1.3.4",
    ],
    "snowflakes": [
        "PasteDeploy==2.1.0",
        "WSGIProxy2==0.4.6",
        "humanfriendly==6.1",
        "pyramid-translogger==0.1",
        "repoze.debug==1.1",
        "rfc3987==1.3.8",
        "rutter==0.2",
        "subprocess_middleware==0.3",
        "waitress==1.4.3",
        "xlrd==1.2.0",
        "zc.buildout==2.13.2",
    ]
}

EXTRAS_REQUIRE["dev"] = EXTRAS_REQUIRE["tests"] + EXTRAS_REQUIRE["snowflakes"]

setup(
    name='snovault',
    version=VERSION,
    description='Snovault Hybrid Object Relational Database Framework',
    long_description=README + CHANGES,
    long_description_content_type="text/x-rst",
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    package_data={'':['nginx-dev.conf']},
    zip_safe=False,
    author='Benjamin Hitz',
    author_email='hitz@stanford.edu',
    url='http://github.com/ENCODE-DCC/snovault/',
    license='MIT',
    install_requires=INSTALL_REQUIRES,
    extras_require=EXTRAS_REQUIRE,
    entry_points='''
        [console_scripts]
        batchupgrade = snovault.batchupgrade:main
        create-mapping = snovault.elasticsearch.create_mapping:main
        dev-servers = snovault.dev_servers:main
        es-index-listener = snovault.elasticsearch.es_index_listener:main

        add-date-created = snowflakes.commands.add_date_created:main
        check-rendering = snowflakes.commands.check_rendering:main
        deploy = snowflakes.commands.deploy:main
        extract_test_data = snowflakes.commands.extract_test_data:main
        es-index-data = snowflakes.commands.es_index_data:main
        import-data = snowflakes.commands.import_data:main
        jsonld-rdf = snowflakes.commands.jsonld_rdf:main
        profile = snowflakes.commands.profile:main
        spreadsheet-to-json = snowflakes.commands.spreadsheet_to_json:main
        migrate-attachments-aws = snowflakes.commands.migrate_attachments_aws:main
        create-admin-user = snowflakes.commands.create_admin_user:main 

        [paste.app_factory]
        main = snowflakes:main
        snowflakes = snowflakes:main

        [paste.composite_factory]
        indexer = snovault.elasticsearch.es_index_listener:composite

        [paste.filter_app_factory]
        memlimit = snowflakes.memlimit:filter_app
        ''',
    classifiers=[
        'Development Status :: 5 - Production/Stable',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Framework :: Pyramid',


        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',
        'Topic :: Database :: Database Engines/Servers',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
    ],
)
